# Generic/Built-in
import json
import copy
import sqlite3
import datetime
import os
import pandas as pd
from typing import List
from hisim.simulationparameters import SimulationParameters
# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils
from dataclasses import dataclass
from  dataclasses_json import dataclass_json
from typing import Any, List, Optional
__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"

@dataclass_json
@dataclass
class EVConfig:
    parameter_string: str
    manufacturer: str
    model: str
    soc: float
    profile: str

    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 manufacturer: str,
                 model: str,
                 soc: float,
                 profile_name: str) -> None:
        self.parameter_string = my_simulation_parameters.get_unique_key()
        self.profile_name = profile_name
        manufacturer = manufacturer
        model = model
        soc= soc


class VehiclePure(cp.Component):
    """
    Vehicle component class

    Parameters
    ----------
    manufacturer : str
        Electric vehicle Manufacturer
    model : str
        Electric vehicle model
    soc : float
        Initial state of charge of battery before
        simulation
    profile : str
        Family profile imported from LPG. The family
        denomination defines the electric vehicle usage
        throughout the simulation duration
    """
    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 manufacturer:str="Tesla",
                 model:str="Model 3 v3",
                 soc:float=1.0,
                 profile:str="CH01") -> None:
        super().__init__(name="EV_charger", my_simulation_parameters=my_simulation_parameters)
        self.evconfig = EVConfig(my_simulation_parameters=my_simulation_parameters,
                                 manufacturer= manufacturer,
                                 model=model,
                                 soc= soc,
                                 profile_name=profile)
        self.build()

    def build(self) -> None:
        """:key

        Defines ...
        max_capacity
        min_capacity
        capacity
        discharing
        car_location
        """
        if self.evconfig.soc > 1 or self.evconfig.soc < 0:
            raise Exception("Invalid State Of Charge.")

        # Gets flexibilities, including heat pump
        electric_vehicle_database = utils.load_smart_appliance("Electric Vehicle")

        electric_vehicle_found = False
        for electric_vehicle in electric_vehicle_database:
            if electric_vehicle["Manufacturer"] == self.evconfig.manufacturer and electric_vehicle["Model"] == self.evconfig.model:
                electric_vehicle_found = True
                break

        if electric_vehicle_found == False:
            raise Exception("Electric Vehicle not registered in the database")

        self.max_capacity = electric_vehicle['Battery Capacity'] * 1E3
        if "Battery Useful Capacity" in electric_vehicle:
            self.min_capacity = self.max_capacity - electric_vehicle["Battery Useful Capacity"] * 1E3
        else:
            self.min_capacity = self.max_capacity * 0.1

        capacity = self.max_capacity * self.evconfig.soc
        if capacity < self.min_capacity:
            capacity = self.min_capacity


        self.capacity = capacity

        cache_file_exists, cache_filepath = utils.get_cache_file(self.component_name, self.evconfig, self.my_simulation_parameters)
        if cache_file_exists:
            self.car_in_charging_station = pd.read_csv(cache_filepath, sep=',', decimal='.')[
                'CarInChargingStation'].tolist()
            self.discharge = pd.read_csv(cache_filepath, sep=',', decimal='.')[
                'Discharge'].tolist()
        else:
            def open_sql(path, table_name):
                sql_file = sqlite3.connect(path)
                return pd.read_sql("SELECT * FROM {};".format(table_name), sql_file)
            def open_ev_json(filepath):
                with open(filepath) as f:
                    data = json.load(f)
                return data["Values"]

            FILEPATH = utils.load_export_load_profile_generator(target=self.evconfig.profile_name)
            if FILEPATH is None:
                FILEPATH = utils.HISIMPATH

            ev_files = dict()
            filepaths = open_sql(FILEPATH["electric_vehicle"][1], "ResultFileEntries")
            list_columns = []
            list_values = []
            for index, row in filepaths.iterrows():
                json_info = json.loads(row["Json"])
                if "Charging" in json_info["FileName"] and "png" not in json_info["FileName"]:
                    filepath = os.path.normpath(json_info["FullFileName"])
                    filepath_list = filepath.split(os.sep)
                    ev_files[filepath_list[-1].split(".")[0]] = json_info["FullFileName"]
                    list_columns.append(filepath_list[-1].split(".")[0])
                    list_values.append(open_ev_json(json_info["FullFileName"]))
            list_values = list(map(list, zip(*list_values)))
            ev_pd = pd.DataFrame(list_values, columns=list_columns)


            # Gets battery information to calculate discharging while not at home
            transportation_devices = open_sql(FILEPATH["electric_vehicle"][0], "TransportationDevices")
            for index, vehicle in transportation_devices.iterrows():
                if "Charging" in vehicle["Name"]:
                    vehicle_info = json.loads(vehicle['Json'])
                    battery_stored_energy_meters = vehicle_info["FullRangeInMeters"]
                    convert_factor = vehicle_info["EnergyToDistanceFactor"]
                    battery_stored_energy_wh = battery_stored_energy_meters * convert_factor


            convert_factor_meters_by_wh = convert_factor * 3600
            soc = []
            load_stats = []
            car_in_charging_station = []
            #car_state = []
            discharge_stats = [0]
            # Gets transportation stats
            transportation_devices_stats = open_sql(FILEPATH["electric_vehicle"][0], "TransportationDeviceStates")
            for index, column in transportation_devices_stats.iterrows():
                if datetime.datetime.strptime(column["DateTime"], '%d/%m/%Y %H:%M') > datetime.datetime.strptime('31/12/2018 23:59', '%d/%m/%Y %H:%M'):
                    if "ParkingAndFullyCharged" in column["DeviceState"] or "ParkingAndCharging" in column["DeviceState"]:
                        car_in_charging_station.append(True)
                    else:
                        car_in_charging_station.append(False)
                    load_stats.append(float(column["CurrentRange"]))
                    soc.append(float(column["CurrentRange"])/battery_stored_energy_meters)
                    if len(load_stats) > 1:
                        diff = load_stats[-1] - load_stats[-2]
                        if diff < 0:
                            discharge_stats.append(diff / convert_factor_meters_by_wh)
                        else:
                            discharge_stats.append(0)

            #discharge = []
            #load = []
            #for index, row in ev_pd.iterrows():
            #    load.append(row["Soc"] * battery_stored_energy_wh)
            #    if index == 0:
            #        discharge.append(0)
            #    else:
            #        diff = load[-1] - load[-2]
            #        if diff < 0:
            #            discharge.append(diff)
            #        else:
            #            discharge.append(0)

            #ev_pd["Discharge"] = discharge
            #data = [ev_pd["CarLocation"].tolist()]
            #data.append(discharge)
            #data.append(car_state)
            #data_parameters = ["CarLocation", "Discharging","CarInChargingStation","RealDischarge","CarState"]

            data:List = []
            data.append(car_in_charging_station)
            data.append(discharge_stats)
            data = list(map(list, zip(*data)))

            data_parameters = ["CarInChargingStation", "Discharge"]
            database = pd.DataFrame(data, columns=data_parameters)

            self.car_in_charging_station = car_in_charging_station
            self.discharge = discharge_stats
            database.to_csv(cache_filepath)
            #utils.save_cache("Vehicle", [self.evconfig.profile_name], database)

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self)  -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool)  -> None:
        pass

class Vehicle(cp.Component):
    """
    Electric Vehicle Component. This is a
    alternative implementation, not fully working,
    to be used in future releases

    Parameters
    ----------
    manufacturer : str
        Electric vehicle Manufacturer
    model : str
        Electric vehicle model
    soc : float
        Initial state of charge of battery before
        simulation
    """
    BeforeCapacity = "BeforeCapacity"

    AfterCapacity = "AfterCapacity"
    MaxCapacity = "MaxCapacity"
    Discharge = "Discharge"

    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 manufacturer:str="Renault",
                 model:str="Zoe v3",
                 soc:float=0.8) -> None:
        super().__init__(name="ElectricVehicle", my_simulation_parameters=my_simulation_parameters)

        self.build(manufacturer=manufacturer, model=model, soc=soc)

        self.before_capacityC: cp.ComponentInput = self.add_input(self.component_name,
                                                                  self.BeforeCapacity,
                                                                  lt.LoadTypes.ELECTRICITY, lt.Units.WATT, True)

        self.after_capacityC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                   self.AfterCapacity,
                                                                   lt.LoadTypes.ANY,
                                                                   lt.Units.WATT)
        self.max_capacityC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                 self.MaxCapacity,
                                                                 lt.LoadTypes.ANY,
                                                                 lt.Units.WATT)
        self.dischargeC : cp.ComponentOutput = self.add_output(self.component_name,
                                                               self.Discharge,
                                                               lt.LoadTypes.ELECTRICITY,
                                                               lt.Units.WATT)

    def build(self, manufacturer: str, model:str, soc:float) -> None:
        if soc > 1 or soc < 0:
            raise Exception("Invalid State Of Charge.")

        # Gets flexibilities, including heat pump
        electric_vehicle_database = utils.load_smart_appliance("Electric Vehicle")

        electric_vehicle_found = False
        for electric_vehicle in electric_vehicle_database:
            if electric_vehicle["Manufacturer"] == manufacturer and electric_vehicle["Model"] == model:
                electric_vehicle_found = True
                break

        if electric_vehicle_found == False:
            raise Exception("Electric Vehicle not registered in the database")

        self.max_capacity = electric_vehicle['Battery Capacity'] * 1E3
        if "Battery Useful Capacity" in electric_vehicle:
            self.min_capacity = self.max_capacity - electric_vehicle["Battery Useful Capacity"] * 1E3
        else:
            self.min_capacity = self.max_capacity * 0.1

        capacity = self.max_capacity * soc
        if capacity < self.min_capacity:
            capacity = self.min_capacity

        self.capacity = capacity

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        if timestep == 0:
            capacity = self.capacity
        else:
            capacity = stsv.get_input_value(self.before_capacityC)
        stsv.set_output_value(self.after_capacityC, capacity)
        stsv.set_output_value(self.max_capacityC, self.max_capacity)

class SimpleStorageState:
    """
    Simplistic implementation for any type
    of energy state storage. Relevant for battery,
    electric vehicles, etc, to store or withdraw
    certain amount of energy of a predefined storage

    Parameters
    ----------
    max_var_val: float
        Maximum variation energy, i.e., maximum power
    min_var_val: float
        Minimum variation energy, i.e., minimum power
    stored_energy : float
        Initial storage energy
    time_correction_factor : float
        Time correction factor to convert power into energy
        the time step duration used in the simulation. This
        factor is equivalent of the invert of seconds per time
        step.
    seconds_per_timestep : int
        Duration in seconds of one time step.
    """
    def __init__(self,
                 max_var_val: float,
                 min_var_val: float,
                 stored_energy: float= 0.0,
                 time_correction_factor:Optional[float]=None,
                 seconds_per_timestep:Optional[int] =None) -> None:
        self.max_var_val = max_var_val
        self.min_var_val = min_var_val
        self.stored_energy = stored_energy
        self.time_correction_factor = time_correction_factor
        self.seconds_per_timestep = seconds_per_timestep

    def store(self, max_capacity: float, current_capacity: float, val: float, efficiency: float = 1.0) -> Any:
        if val < max_capacity - current_capacity:
            amount = val
        else:
            amount = max_capacity - current_capacity

        # Check if the theoretical capacity is larger than maximum capacity
        if current_capacity >= max_capacity:
            amount = 0

        # Check if the charging variation is larger than physical possible
        if amount > self.max_var_val:
            amount = self.max_var_val

        amount = amount * efficiency

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def force_store(self, max_capacity: float, current_capacity: float) -> Any:
        amount = self.max_var_val
        if amount > max_capacity - current_capacity:
            amount = max_capacity - current_capacity

        # Check if the theoretical capacity is larger than maximum capacity
        if current_capacity >= max_capacity:
            amount = 0

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def withdraw(self, min_capacity: float, current_capacity: float, val: float, efficiency: float = 1.0) -> Any:
        val = abs(val)
        if current_capacity - min_capacity > val:
            amount = -val
        else:
            amount = min_capacity - current_capacity

        if current_capacity <= min_capacity:
            amount = 0

        if amount < self.min_var_val:
            amount = self.min_var_val

        amount = amount / efficiency

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def keep_state(self, capacity: float) -> Any:
        charging_delta = 0
        after_capacity = capacity
        return charging_delta, capacity

class EVCharger(cp.Component):
    """
    Electric Vehicle Charger Component
    Parameters:
    ----------------
        manufacturer : str
        name : str
        electric_vehicle : Vehicle_Pure
        sim_params : cp.SimulationParameters

    """
    # Inputs
    ElectricityInput = "ElectricityInput"
    #StoredEnergy = "StoredEnergy"
    #MaxStoredEnergy = "MaxStoredEnergy"
    EVChargerState = "EVChargerState"
    EVChargerMode = "EVChargerMode"
    MinimumStateOfCharge = "MinimumStateOfCharge"

    # Outputs
    StateOfCharge = "StateOfCharge"
    AfterStoredEnergy = "AfterStoredEnergy"
    Driving = "Driving"
    AtChargingStation = "AtChargingStation"
    ElectricityOutput = "ElectricityOutput"

    # Similar components to connect to:
    # 1. EVChargerController
    # 2. Some ChargingInput

    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 manufacturer:str="myenergi",
                 name:str="Wallbox ZAPPI 222TW",
                 electric_vehicle:Any=None ) -> None:
        super().__init__(name="EVCharger", my_simulation_parameters=my_simulation_parameters)

        self.build(manufacturer=manufacturer, name=name, electric_vehicle=electric_vehicle, sim_params=my_simulation_parameters)

        self.state = SimpleStorageState(max_var_val=self.charging_power,
                                        min_var_val=-self.charging_power,
                                        stored_energy=self.electric_vehicle.capacity,
                                        time_correction_factor=self.time_correction_factor,
                                        seconds_per_timestep=self.seconds_per_timestep)


        self.previous_state = copy.deepcopy(self.state)

        self.charging_inputC : cp.ComponentInput = self.add_input(self.component_name,
                                                                  self.ElectricityInput,
                                                                  lt.LoadTypes.ELECTRICITY,
                                                                  lt.Units.WATT,
                                                                  True)
        self.stateC: cp.ComponentInput = self.add_input(self.component_name,
                                                        self.EVChargerState,
                                                        lt.LoadTypes.ANY,
                                                        lt.Units.ANY,
                                                        True)
        self.modeC: cp.ComponentInput = self.add_input(self.component_name,
                                                       self.EVChargerMode,
                                                       lt.LoadTypes.ANY,
                                                       lt.Units.ANY,
                                                       True)
        self.min_socC: cp.ComponentInput = self.add_input(self.component_name,
                                                          self.MinimumStateOfCharge,
                                                          lt.LoadTypes.ANY,
                                                          lt.Units.ANY,
                                                          True)

        self.socC: cp.ComponentOutput = self.add_output(self.component_name,
                                                        self.StateOfCharge,
                                                        lt.LoadTypes.ANY,
                                                        lt.Units.ANY)
        self.after_capacityC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                   self.AfterStoredEnergy,
                                                                   lt.LoadTypes.ANY,
                                                                   lt.Units.ANY)
        self.drivingC : cp.ComponentOutput = self.add_output(self.component_name,
                                                             self.Driving,
                                                             lt.LoadTypes.ANY,
                                                             lt.Units.ANY)
        self.at_charging_stationC : cp.ComponentOutput = self.add_output(self.component_name,
                                                                         self.AtChargingStation,
                                                                         lt.LoadTypes.ANY,
                                                                         lt.Units.ANY)
        self.electricity_outputC : cp.ComponentOutput = self.add_output(self.component_name,
                                                                        self.ElectricityOutput,
                                                                        lt.LoadTypes.ELECTRICITY,
                                                                        lt.Units.WATT,
                                                                        sankey_flow_direction=True)

    def build(self, manufacturer: str, name:str, electric_vehicle: Any, sim_params:SimulationParameters) -> None:
        self.time_correction_factor = 1 / sim_params.seconds_per_timestep
        self.seconds_per_timestep = sim_params.seconds_per_timestep

        electric_vehicle_charger_database = utils.load_smart_appliance("Wallbox")
        evcharger_found = False
        for evcharger in electric_vehicle_charger_database:
            if evcharger["Manufacturer"] == manufacturer and evcharger["Name"] == name:
                charging_power = evcharger["Charging Power"] * 1E3
                evcharger_found = True
                break

        if evcharger_found == False:
            raise Exception("EV Charging Station has not been found in the database.")

        self.manufacturer = manufacturer
        self.name = name
        self.charging_power_original = charging_power * 1E-3
        self.charging_power = charging_power * self.time_correction_factor
        self.electric_vehicle = electric_vehicle

    def write_to_report(self) -> List[str]:
        lines = []
        lines.append("Name: {}".format(self.component_name))
        lines.append("Manufacturer: {}".format(self.manufacturer))
        lines.append("Model: {}".format(self.name))
        lines.append("Charging Power: {} kW".format(self.charging_power_original))
        lines.append("EV Battery Capacity: {}".format(self.electric_vehicle.max_capacity*1E-3))
        lines.append("Vehicle: {}".format(self.electric_vehicle.model))
        return lines

    def i_save_state(self) -> None:
        self.previous_state = copy.deepcopy(self.state)

    def i_restore_state(self) -> None:
        self.state = copy.deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        if force_convergence:
            return

        # Gets inputs
        charging = stsv.get_input_value(self.charging_inputC)
        state = stsv.get_input_value(self.stateC)
        mode = stsv.get_input_value(self.modeC)
        min_soc = stsv.get_input_value(self.min_socC)

        capacity = self.state.stored_energy
        max_capacity = self.electric_vehicle.max_capacity

        min_capacity = max_capacity * min_soc

        if bool(self.electric_vehicle.car_in_charging_station[timestep]):
            connected_to_charging_station = 1
            driving = 0
            to_be_charged = -charging

            if state == 0:
                charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
            elif state == 1:
                charging_delta, after_capacity = self.charge_only_from_the_grid(max_capacity=max_capacity,
                                                                                capacity=capacity)
            elif state == 2:
                charging_delta, after_capacity = self.charge_only_electricity_surplus(to_be_charged=to_be_charged,
                                                                                      max_capacity=max_capacity,
                                                                                      capacity=capacity)
            elif state == 3:
                charging_delta, after_capacity = self.operate_on_vehicle_to_grid(to_be_charged=to_be_charged,
                                                                                 max_capacity=max_capacity,
                                                                                 min_capacity=min_capacity,
                                                                                 capacity=capacity)
            else:
                raise Exception("State {} has not been implemented! "
                                "Please check EVCharger and EVChargerController".format(state))

        # Discharges battery according to electric vehicle use
        elif self.electric_vehicle.discharge[timestep] != 0.0:
            connected_to_charging_station = 0
            driving = 1

            charging_delta, after_capacity = self.state.withdraw(min_capacity=0,
                                                                 current_capacity=capacity,
                                                                 val=self.electric_vehicle.discharge[timestep])
            charging_delta = 0
        else:
            connected_to_charging_station = 0
            driving = 0
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)


        stsv.set_output_value(self.socC, after_capacity/max_capacity)
        stsv.set_output_value(self.after_capacityC, after_capacity)
        stsv.set_output_value(self.drivingC, driving)
        stsv.set_output_value(self.electricity_outputC, charging_delta)
        stsv.set_output_value(self.at_charging_stationC, connected_to_charging_station)

    def charge_only_electricity_surplus(self, to_be_charged, max_capacity, capacity):
        if to_be_charged >= 0:
            charging_delta, after_capacity = self.state.store(max_capacity=max_capacity,
                                                              current_capacity=capacity,
                                                              val=to_be_charged)
        else:
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
        return charging_delta, after_capacity

    def operate_on_vehicle_to_grid(self, to_be_charged: float, max_capacity: float, min_capacity: float, capacity: float) -> Any:
        if to_be_charged >= 0:
            charging_delta, after_capacity = self.state.store(max_capacity=max_capacity,
                                                              current_capacity=capacity,
                                                              val=to_be_charged)
        elif to_be_charged < 0:
            charging_delta, after_capacity = self.state.withdraw(min_capacity=min_capacity,
                                                                 current_capacity=capacity,
                                                                 val=to_be_charged)
        else:
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
        return charging_delta, after_capacity

    def charge_only_from_the_grid(self, max_capacity, capacity):
        charging_delta, after_capacity = self.state.force_store(max_capacity=max_capacity, current_capacity=capacity)
        return charging_delta, after_capacity

class EVChargerController(cp.Component):
    """
    Imports data from Load Profile Generator and
    uses as a Component Class in hisim
    """
    # Inputs
    ElectricityInput = "ElectricityInput"
    StateOfCharge = "StateOfCharge"

    # Outputs
    EVChargerState = "EVChargerState"
    EVChargerMode = "EVChargerMode"
    MinimumStateOfCharge = "MinimumStateOfCharge"

    # Similar components to connect to:
    # 1. Some ChargingInput

    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 mode:int=1) -> None:
        super().__init__(name="EVChargerController", my_simulation_parameters=my_simulation_parameters)
        self.mode = mode

        if mode == 1:
            self.mode_description = "Straight Charging"
            self.mode_extended_description = "Charge the Electric Vehicle whenever is connected to EV Charger"
        elif mode == 2:
            self.mode_description = "Charge only on Electricity Surplus"
            self.mode_extended_description = "Charge the Electric Vehicle whenever the home grid is on electricity surplus"
        elif mode == 3:
            self.mode_description = "Operate on Vehicle-to-Grid"
            self.mode_extended_description = "Charge the Electric Vehicle whenever the home grid is on electricity surplus. " \
                                             "Discharge the Electric Vehicle whenenver the home grid is running on electricity deficit."
        elif mode == 4:
            self.mode_description = "Stepped Prioritized Charging"
            self.mode_extended_description = "Charge only from the grid for SoC from 0% up to 60%. " \
                                             "Charge only from the grid only on the electricity surplus for SoC from 60% up to 80%. " \
                                             "Operate on Vehicle-to-Grid for SoC from 80% up to 100%"
        elif mode == 5:
            self.mode_description = "Tight Stepped Prioritized Charging"
            self.mode_extended_description = "Charge only from the grid for SoC from 0% up to 40%. " \
                                             "Charge only from the grid only on the electricity surplus for SoC from 40% up to 70%. " \
                                             "Operate on Vehicle-to-Grid for SoC from 70% up to 100%"
        elif mode == 6:
            self.mode_description = "Super Tight Stepped Prioritized Charging"
            self.mode_extended_description = "Charge only from the grid for SoC from 0% up to 20%. " \
                                         "Charge only from the grid only on the electricity surplus for SoC from 20% up to 60%. " \
                                         "Operate on Vehicle-to-Grid for SoC from 60% up to 100%"
        else:
            self.mode_description = "WRITE MODE DESCRIPTION HERE!"
            self.mode_extended_description = "WRITE MODE EXTENDED DESCRIPTION HERE!"

        # Inputs
        self.charging_inputC: cp.ComponentInput        = self.add_input(self.component_name,
                                                                        self.ElectricityInput,
                                                                        lt.LoadTypes.ELECTRICITY,
                                                                        lt.Units.WATT,
                                                                        True)
        self.socC: cp.ComponentInput = self.add_input(self.component_name,
                                                      self.StateOfCharge,
                                                      lt.LoadTypes.ANY,
                                                      lt.Units.ANY,
                                                      True)


        self.stateC: cp.ComponentOutput = self.add_output(self.component_name,
                                                          self.EVChargerState,
                                                          lt.LoadTypes.ANY,
                                                          lt.Units.ANY)
        self.modeC: cp.ComponentOutput = self.add_output(self.component_name,
                                                         self.EVChargerMode,
                                                         lt.LoadTypes.ANY,
                                                         lt.Units.ANY)
        self.min_socC: cp.ComponentOutput = self.add_output(self.component_name,
                                                            self.MinimumStateOfCharge,
                                                            lt.LoadTypes.ANY,
                                                            lt.Units.ANY)

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues)  -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        # Gets inputs
        charging = stsv.get_input_value(self.charging_inputC)
        state_of_charge = stsv.get_input_value(self.socC)

        minimum_state_of_charge = 0.1

        # Straight Charging
        if self.mode == 1:
            state = 1
        # Charge only on Electricity Surplus
        elif self.mode == 2:
            state = 2
        # Operate on Vehicle-to-Grid
        elif self.mode == 3:
            state = 3
        # Stepped Prioritized Charging
        elif self.mode == 4:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.6:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif state_of_charge >= 0.6 and state_of_charge < 0.8:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.8:
                minimum_state_of_charge = 0.8
                state = 3

        # Tight Stepped Prioritized Charging
        elif self.mode == 5:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.4:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif state_of_charge >= 0.4 and state_of_charge < 0.7:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.7:
                minimum_state_of_charge = 0.7
                state = 3

        # Super Tight Stepped Prioritized Charging
        elif self.mode == 6:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.2:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif state_of_charge >= 0.2 and state_of_charge < 0.6:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.6:
                minimum_state_of_charge = 0.6
                state = 3
        else:
            if self.mode is None:
                raise Exception("No mode has been implemented.")
            else:
                raise Exception("Mode {} has not been implemented yet.".format(self.mode))

        stsv.set_output_value(self.stateC, state)
        stsv.set_output_value(self.min_socC, minimum_state_of_charge)
        stsv.set_output_value(self.modeC, self.mode)


    def write_to_report(self):
        lines = []
        lines.append("Mode Number: {}".format(self.mode))
        lines.append("Mode: {}".format(self.mode_description))
        try:
            lines.append("Extend Mode Description: {}".format(self.mode_extended_description))
        except AttributeError:
            lines.append("Extend Mode Description: TO BE IMPLEMENTED")
        return lines



""" Charts for a single day. """
# clean
import os
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter

from hisim.postprocessing.chartbase import Chart


class ChartSingleDay(Chart):

    """ For making visualisations for a single day. """

    def __init__(self, output, units, directorypath, time_correction_factor, data, day=None, month=None, output2=None):
        """ Initializes the class. """
        if output2 is not None:
            super().__init__(output, "days", units, directorypath, time_correction_factor, output2)
        else:
            super().__init__(output, "days", units, directorypath, time_correction_factor)
        self.axis: plt.axis
        self.ax2: plt.axis
        self.line2: plt.axis
        self.month = month
        self.day = day
        self.data = data
        self.plot_title: str
        self.filename = f"{self.type.lower()}_{self.output.split(' # ', 2)[1]}_{self.output.split(' # ', 2)[0]}_m" \
                        f"{self.month}_d{self.day}.png"
        self.filepath = os.path.join(self.directorypath, self.filename)

    def get_day_data(self):
        """ Extracts data for a single day. """
        firstindex = (self.month * 30 + self.day) * 24 * int(1 / self.time_correction_factor)
        lastindex = firstindex + 24 * int(1 / self.time_correction_factor)
        day_number = self.day + 1
        if day_number == 1:
            ordinal = "st"
        elif day_number == 2:
            ordinal = "nd"
        elif day_number == 3:
            ordinal = "rd"
        else:
            ordinal = "th"
        date = f"{self.label_months_lowercase[self.month]} {day_number}{ordinal}"
        self.plot_title = f"{self.title} {date}"

        if abs(lastindex - firstindex) < len(self.data):
            data = self.data[firstindex:lastindex]
            data_index = data.index[firstindex:lastindex]
            data.index = data_index
            return data
        return self.data

    def __add__(self, other):
        """ Adds another chart to this one. """
        my_double: ChartSingleDay = ChartSingleDay(self.output, self.ylabel, self.directorypath, self.time_correction_factor, self.day, self.month)
        my_double.filename = f"{self.type.lower()}_{self.output.split(' # ', 2)[1]}_{self.output.split(' # ', 2)[0]}" \
                             f"_AND_{other.output.split(' # ', 2)[1]}_{other.output.split(' # ', 2)[0]}_m{self.month}_d{self.day}.png"
        my_double.filepath = os.path.join(self.directorypath, my_double.filename)
        my_double.plot(close=False)

        #  twin object for two different y-axis on the sample plot
        my_double.ax2 = my_double.axis.twinx()
        #  make a plot with different y-axis using second axis object
        my_double.line2 = my_double.ax2.plot(self.data.index, other.data, label=other.property, linewidth=5)
        my_double.ax2.set_ylabel(f"{other.property} [{other.ylabel}]", fontsize=18)
        my_double.ax2.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        return my_double

    def close(self):
        """ Closes a chart and saves. """
        plt.xticks(fontsize=18)
        plt.yticks(fontsize=18)
        self.axis.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        self.axis.set_ylabel(f"{self.property} [{self.ylabel}]", fontsize=18)
        if hasattr(self, "line2"):
            self.ax2.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        plt.savefig(self.filepath)
        plt.close()

    def plot(self, close):
        """ Plots a chart. """
        single_day_data = self.get_day_data()
        plt.xticks(fontsize=18)
        plt.yticks(fontsize=18)
        plt.rcParams['font.size'] = '18'
        plt.rcParams['agg.path.chunksize'] = 10000
        _fig, self.axis = plt.subplots(figsize=(13, 9))
        plt.xticks(fontsize=18)
        plt.yticks(fontsize=18)
        if abs(max(single_day_data)) > 1.5E3:
            single_day_data = single_day_data * 1E-3
            self.ylabel = f"k{self.ylabel}"
        plt.plot(single_day_data.index, single_day_data, color="green", linewidth=5.0, label=self.property)
        plt.grid(True)
        self.axis.set_ylabel(self.ylabel)
        plt.xlabel("Time [hours]", fontsize=18)  # fontsize=18)
        self.axis.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        if close:
            self.close()



# Import packages from standard library or the environment e.g. pandas, numpy etc.
from typing import List, Any
from copy import deepcopy
from dataclasses import dataclass
from bslib import bslib as bsl
from dataclasses_json import dataclass_json

# Import modules from HiSim
from hisim.component import Component, ComponentInput, ComponentOutput, SingleTimeStepValues
from hisim.loadtypes import LoadTypes, Units, InandOutputType, ComponentType
from hisim.simulationparameters import SimulationParameters
from typing import Optional
from hisim.components import controller_l1_generic_ev_charge

__authors__ = "Tjarko Tjaden, Hauke Hoops, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = "..."
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Tjarko Tjaden"
__email__ = "tjarko.tjaden@hs-emden-leer.de"
__status__ = "development"

@dataclass_json
@dataclass
class BatteryConfig:
    system_id: str
    p_inv_custom: float  # power in Watt
    e_bat_custom: float  # capacity in Kilowatt
    name: str
    source_weight : int

    @staticmethod
    def get_default_config(name: str = 'Battery', p_inv_custom: float = 5, e_bat_custom: float = 10, source_weight: int = 1) -> Any:
        config=BatteryConfig(
            system_id='SG1',
            p_inv_custom=p_inv_custom,
            e_bat_custom=e_bat_custom,
            name=name,
            source_weight=source_weight)
        return config

class Battery(Component):
    """
    Simulate state of charge and realized power of a ac coupled battery
    storage system with the bslib library. Relevant simulation parameters
    are loaded within the init for a specific or generic battery type.
    """

    # Inputs
    LoadingPowerInput = "LoadingPowerInput"     # W

    # Outputs
    AcBatteryPower = "AcBatteryPower"           # W
    DcBatteryPower = "DcBatteryPower"           # W
    StateOfCharge = "StateOfCharge"             # [0..1]

    def __init__(self, my_simulation_parameters: SimulationParameters,
                 config:BatteryConfig):
        """
        Loads the parameters of the specified battery storage.

        Parameters
        ----------
        system_id : str
            Name (system_id) of the battery storage from bslib database.
        p_inv_custom : numeric, default 0
            AC power of battery inverter. Only for system_ids of type "Generic". [W]
        e_bat_custom : numeric, default 0
            Useable battery capacity. Only for system_ids of type "Generic". [Wh]
        """
        self.battery_config = config
        super().__init__(name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters)

        self.source_weight = self.battery_config.source_weight

        self.system_id = self.battery_config.system_id

        self.p_inv_custom = self.battery_config.p_inv_custom

        self.e_bat_custom = self.battery_config.e_bat_custom

        # Component has states
        self.state = BatteryState()
        self.previous_state = deepcopy(self.state)

        # Load battery object with parameters from bslib database
        self.BAT = bsl.ACBatMod(system_id=self.system_id,
                                p_inv_custom=self.p_inv_custom,
                                e_bat_custom=self.e_bat_custom)

        # Define component inputs
        self.p_set: ComponentInput = self.add_input(object_name=self.component_name,
                                                    field_name=self.LoadingPowerInput,
                                                    load_type=LoadTypes.ELECTRICITY,
                                                    unit=Units.WATT,
                                                    mandatory=True)

        # Define component outputs
        self.p_bs: ComponentOutput = self.add_output(object_name=self.component_name,
                                                     field_name=self.AcBatteryPower,
                                                     load_type=LoadTypes.ELECTRICITY,
                                                     unit=Units.WATT,
                                                     postprocessing_flag=[InandOutputType.CHARGE_DISCHARGE, ComponentType.BATTERY])

        self.p_bat: ComponentOutput = self.add_output(object_name=self.component_name,
                                                      field_name=self.DcBatteryPower,
                                                      load_type=LoadTypes.ELECTRICITY,
                                                      unit=Units.WATT)

        self.soc: ComponentOutput = self.add_output(object_name=self.component_name,
                                                    field_name=self.StateOfCharge,
                                                    load_type=LoadTypes.ANY,
                                                    unit=Units.ANY,
                                                    postprocessing_flag=[InandOutputType.STORAGE_CONTENT])


    def i_save_state(self)  -> None:
        self.previous_state = deepcopy(self.state)

    def i_restore_state(self)  -> None:
        self.state = deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int,  stsv: SingleTimeStepValues) -> None:
        pass
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues,  force_convergence: bool)  -> None:

        # Parameters
        dt = self.my_simulation_parameters.seconds_per_timestep

        # Load input values
        p_set = stsv.get_input_value(self.p_set)
        soc = self.state.soc

        # Simulate on timestep
        results = self.BAT.simulate(p_load=p_set,
                                    soc=soc,
                                    dt=dt)
        p_bs = results[0]
        p_bat = results[1]
        soc = results[2]

        # write values for output time series
        stsv.set_output_value(self.p_bs, p_bs)
        stsv.set_output_value(self.p_bat, p_bat)
        stsv.set_output_value(self.soc, soc)

        # write values to state
        self.state.soc = soc

    def write_to_report(self) -> List[str]:
        lines = []
        lines.append("Advanced Battery bslib: " + self.component_name)
        return lines


@dataclass
class BatteryState:
    soc: float = 0



""" Data Transfer Object to get all the result data to the post processing. """

# clean
from typing import Any, List

from hisim import log
from hisim.component_wrapper import ComponentWrapper
from hisim.simulationparameters import SimulationParameters


class PostProcessingDataTransfer:  # noqa: too-few-public-methods

    """ Data class for transfering the result data to this class. """

    def __init__(self,
                 results: Any,
                 all_outputs: Any,
                 simulation_parameters: SimulationParameters,
                 wrapped_components: List[ComponentWrapper],
                 mode: Any,
                 setup_function: Any,
                 execution_time: Any,
                 results_monthly: Any,
                 ) -> None:
        """ Initializes the values. """
        # Johanna Ganglbauer: time correction factor is applied in postprocessing to sum over power values and convert them to energy
        self.time_correction_factor = simulation_parameters.seconds_per_timestep / 3600
        self.results = results
        self.all_outputs = all_outputs
        self.simulation_parameters = simulation_parameters
        self.wrapped_components: List[ComponentWrapper] = wrapped_components
        self.mode = mode
        self.setup_function = setup_function
        self.execution_time = execution_time
        self.results_monthly = results_monthly
        self.post_processing_options = simulation_parameters.post_processing_options

        log.information("Selected " + str(len(self.post_processing_options)) + " post processing options:")
        for option in self.post_processing_options:
            log.information("Selected post processing option: " + str(option))



# Owned
from typing import  List
from hisim import component as cp
from hisim.simulationparameters import SimulationParameters

from hisim import loadtypes as lt

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


class PriceSignal(cp.Component):
    """
    Class component that provides price for electricity.

    Parameters
    -----------------------------------------------
    profile: string
        profile code corresponded to the family or residents configuration

    ComponentInputs:
    -----------------------------------------------
       None

    ComponentOutputs:
    -----------------------------------------------
       Price for injection: cents/kWh
       Price for purchase: cents/kWh
    """
    #Forecasts
    Price_Injection_Forecast_24h = "Price_Injection_Forecast_24h"
    Price_Purchase_Forecast_24h = "Price_Purchase_Forecast_24h"
    
    #Outputs
    PricePurchase = 'PricePurchase'
    PriceInjection = 'PriceInjection'

    def __init__( self,
                  my_simulation_parameters: SimulationParameters ) -> None:
        super( ).__init__( name = "PriceSignal", my_simulation_parameters = my_simulation_parameters )

        self.build_dummy( start = int( 10 * 3600 / my_simulation_parameters.seconds_per_timestep ), 
                          end = int( 16 * 3600 / my_simulation_parameters.seconds_per_timestep ) )
        
        # Outputs
        self.PricePurchaseC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                  self.PricePurchase,
                                                                  lt.LoadTypes.PRICE,
                                                                  lt.Units.EUR_PER_KWH,
                                                                  postprocessing_flag = [
                                                                    lt.LoadTypes.PRICE,
                                                                    lt.InandOutputType.ELECTRICITY_CONSUMPTION
                                                                    ]
                                                                    )
        self.PriceInjectionC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                   self.PriceInjection,
                                                                   lt.LoadTypes.PRICE,
                                                                   lt.Units.EUR_PER_KWH,
                                                                   postprocessing_flag=[
                                                                    lt.LoadTypes.PRICE,
                                                                    lt.InandOutputType.ELECTRICITY_INJECTION
                                                                    ]
                                                                    )

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_conversion: bool) -> None:
        if self.my_simulation_parameters.system_config.predictive and self.my_simulation_parameters.system_config.prediction_horizon:
            priceinjectionforecast = [0.1] * int(self.my_simulation_parameters.system_config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep)
            pricepurchaseforecast = [0.5] * int(self.my_simulation_parameters.system_config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep)
        else:
            priceinjectionforecast = [0.1]
            pricepurchaseforecast = [0.5]
        
        self.simulation_repository.set_entry(self.Price_Injection_Forecast_24h, priceinjectionforecast)
        self.simulation_repository.set_entry(self.Price_Purchase_Forecast_24h, pricepurchaseforecast)
        stsv.set_output_value(self.PricePurchaseC, pricepurchaseforecast[0])
        stsv.set_output_value(self.PriceInjectionC, priceinjectionforecast[0])

    def build_dummy( self, start : int, end: int ) -> None:
        self.start = start
        self.end = end

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def write_to_report(self) -> List[str]:
        lines = []
        lines.append( "Price signal: {}".format( "dummy" ) )
        return lines



""" Contains all the chart classes. """
# clean
import gc
from typing import Any
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from hisim import log
from hisim.postprocessing.chartbase import Chart
from hisim import utils
from hisim.components import building
from hisim.components import generic_heat_pump_modular
from hisim.components import generic_heat_pump
from hisim.components import advanced_heat_pump_hplib
from hisim.components import loadprofilegenerator_connector
mpl.rcParams['agg.path.chunksize'] = 10000


class Carpet(Chart):  # noqa: too-few-public-methods

    """ Class for carpet plots. """

    def __init__(self, output: Any, units: Any, directorypath: str, time_correction_factor: float) -> None:
        """ Initalizes a carpot plot. """
        super().__init__(output=output,
                         chart_type="Carpet",
                         units=units,
                         directory_path=directorypath,
                         time_correction_factor=time_correction_factor)

    def plot(self, xdims: int, data: Any) -> None:
        """ Makes a carpet plot. """
        log.trace("starting carpet plots")
        ydims = int(len(data) / xdims)  # number of calculated timesteps per day
        y_steps_per_hour = int(ydims / 24)
        try:
            database = data.values.reshape(xdims, ydims)
        except ValueError:
            log.error("Carpet plot can only deal with data containing entire days")
            return
        if np.max(np.abs(data.values)) > 1.5E3:
            database = database * 1E-3
            self.units = f"k{self.units}"
        plot_data = np.flip(database.transpose(), axis=0)

        fig = plt.figure(figsize=(10, 5), dpi=500)

        axis = fig.add_subplot(111)
        mycolors = 'viridis'
        color_map = plt.cm.get_cmap(mycolors)

        plot = axis.pcolormesh(plot_data, cmap=color_map)
        plt.colorbar(plot).set_label(self.units)

        y_ticks = np.arange(0, 25 * y_steps_per_hour, 6 * y_steps_per_hour).tolist()
        axis.set_yticks(y_ticks)
        y_ticks_labels = np.flip(list(range(0, 25, 6)), axis=0)
        axis.set_yticklabels([str(i) for i in y_ticks_labels])

        if xdims == 365:
            x_ticks = np.arange(15, 346, 30).tolist()
            axis.set_xticks(x_ticks)
            axis.set_xticklabels([str(i) for i in self.months_abbrev_uppercase])

        # optimizing fonts
        fig.autofmt_xdate(rotation=45)
        # setting axis of the plot
        axis.set_ylabel('Daytime [h]')
        axis.set_xlabel('Month of the year')

        log.trace("finished carpet plot: " + self.filepath)
        plt.savefig(self.filepath, bbox_inches='tight')
        plt.close()


class Line(Chart):  # noqa: too-few-public-methods

    """ Makes a line chart. """

    # @utils.measure_memory_leak
    def __init__(self, output, units, directorypath, time_correction_factor):
        """ Initializes a line chart. """
        super().__init__(output, "line", units, directorypath, time_correction_factor)

    @utils.measure_memory_leak
    def plot(self, data, units):
        """ Makes a line plot. """
        all_font_size = 40
        size_1 = 20
        size_2 = 18
        mpl.use('Agg')

        ylabel = units
        _fig, axis = plt.subplots(figsize=(size_1, size_2))
        x_zero = data.index
        plt.xticks(fontsize=35, rotation=20)
        plt.yticks(fontsize=35)

        # Rescale values in case they are too high
        if max(abs(data)) > 1.5E3 and units != "-":
            data = data * 1E-3
            units = f"k{units}"

        plt.plot(x_zero, data, color="green", linewidth=6.0)
        plt.ylabel(ylabel, fontsize=all_font_size)
        plt.ylabel(f"[{units}]", fontsize=40)
        plt.xlabel("Time", fontsize=all_font_size)
        plt.grid()
        axis.set_xlim(xmin=x_zero[0])
        plt.savefig(self.filepath)
        plt.cla()
        plt.clf()
        plt.close("all")
        del x_zero
        gc.collect(2)


class BarChart(Chart):  # noqa: too-few-public-methods

    """ Makes Bar charts. """

    original = [385.66, 484.01, 981.05, 1096.7, 1157, 1299.9, 1415.3, 1266.1, 1075.8, 714.44, 422.51, 366.83]

    def __init__(self, output, units, dirpath, time_correction_factor):
        """ Initializes the classes. """
        super().__init__(output, "Bar", units, dirpath, time_correction_factor)
        self.filename = f"monthly_{self.output}.png"

    def plot(self, data):
        """ Plots the bar chart. """
        width = 0.35
        # Specify the values of blue bars (height)

        # Position of bars on x-axis
        ind = np.arange(12)

        # Figure size

        # Width of a bar
        width = 0.4

        plt.subplots()
        plt.bar(ind, data * 1E-3, width, label="HiSim")
        plt.bar(ind + width, self.original, width, label="PVSOL")

        plt.xticks(ind + width / 2)

        plt.title(f"{self.title} Monthly")
        plt.grid()
        plt.tight_layout()
        plt.ylabel(self.units)
        plt.legend(loc='best')
        plt.savefig(self.filepath, bbox_inches='tight')
        plt.close()


class SankeyHISIM(Chart):

    """ Class for sankey charts. """

    def __init__(self,
                 name,
                 units,
                 directorypath,
                 time_correction_factor):
        """ Initializes the Sankey chart. """
        super().__init__(output=name,
                         chart_type="Sankey",
                         units=units,
                         directory_path=directorypath,
                         time_correction_factor=time_correction_factor)
        self.filename = f"{self.output}.png"

    def plot(self, data):
        """ Executes the plot. """
        components = {}
        for _index, output_result in enumerate(data):
            if self.output == output_result.display_name:
                if output_result.sankey_flow_direction is True:
                    components[output_result.component_name] = round(sum(output_result.Results) * 1E-3)
                elif output_result.sankey_flow_direction is False:
                    components[output_result.component_name] = - round(sum(output_result.Results) * 1E-3)

        # if components:
        flows = []
        flows_labels = []
        for key in components.items():
            if isinstance(components[key], bool) is False:
                flows.append(components[key])
                flows_labels.append(f"{key} kWh")

        orientations = self.make_orientations(flows)

        pathlengths = 0.4
        fig = plt.figure(figsize=[10, 10])
        axis = fig.add_subplot(1, 1, 1, xticks=[], yticks=[])

        sankey = mpl.sankey.Sankey(ax=axis, scale=5E-5, offset=3E-1, head_angle=100, margin=0.4)
        sankey.add(flows=flows,
                   labels=flows_labels,
                   orientations=orientations,
                   pathlengths=pathlengths)
        sankey.finish()
        plt.title(self.title, fontsize=18)
        plt.axis("off")
        plt.savefig(self.filepath)
        plt.close()

    def make_orientations(self, flows):
        """ Counts the orientations. """
        i_positive = 0
        i_negative = 0
        indices = [0, 1, -1]
        orientations = []
        for _index, flow in enumerate(flows):
            if flow > 0:
                orientations.append(indices[i_positive])
                i_positive = i_positive + 1
            else:
                orientations.append(indices[i_negative])
                i_negative = i_negative + 1
            if i_positive >= 3:
                i_positive = 0
            if i_negative >= 3:
                i_negative = 0
        return orientations

    def plot_heat_pump(self, data):
        """ Plots a sankey for the Heat Pump. """
        electricity_consumption = 0

        for _index, output_result in enumerate(data):
            if output_result.component_name == "HeatPump":
                if output_result.display_name == "ElectricityOutput":
                    electricity_consumption = sum(output_result.Results)
                if output_result.display_name == "Heating":
                    heating = sum(output_result.Results)
                if output_result.display_name == "Cooling":
                    cooling = sum(output_result.Results)
        if cooling < 0:
            cooling = abs(cooling)
        thermal_energy_delivered = heating + cooling
        thermal_energy_extracted = thermal_energy_delivered - electricity_consumption

        flows = [electricity_consumption * 1E-3,
                 thermal_energy_extracted * 1E-3,
                 -thermal_energy_delivered * 1E-3]
        flows = [round(i) for i in flows]
        flows_labels = ['Electricity\nConsumption',
                        'Thermal Energy\nExtracted',
                        'Thermal Energy\nDelivered']
        orientations = [1, -1, 0]
        pathlengths = 0.25

        fig = plt.figure(figsize=[10, 10])
        axis = fig.add_subplot(1, 1, 1, xticks=[], yticks=[])

        sankey = mpl.sankey.Sankey(ax=axis, scale=5E-5, offset=3E-1, head_angle=100, margin=0.4)
        sankey.add(flows=flows,
                   labels=flows_labels,
                   orientations=orientations,
                   pathlengths=pathlengths)
        sankey.finish()
        plt.title("Heap Pump Energy Equilibrium", fontsize=18)
        plt.axis("off")
        plt.savefig(self.filepath)
        plt.close()

    def plot_building(self, data):
        """ Sankey of Thermal Equilibrium from Residence. """
        heating_by_residents, internal_loss, solar_gain_through_windows, \
            stored_energy_variation, thermal_energy_delivered, total_energy_to_residence = self.calculate_metrics(data)

        # If the
        if abs(stored_energy_variation) < 100 * 1E3:
            flows = [heating_by_residents * 1E-3,
                     solar_gain_through_windows * 1E-3,
                     thermal_energy_delivered * 1E-3,
                     -internal_loss * 1E-3]
            flows = [round(i) for i in flows]
            flows_labels = ['Heating\nby Residents',
                            'Solar\nGain',
                            'Heater',
                            'Heat\nLoss']
            orientations = [1, 0, -1, 0]
        else:

            flows = [heating_by_residents * 1E-3,
                     solar_gain_through_windows * 1E-3,
                     thermal_energy_delivered * 1E-3,
                     -stored_energy_variation * 1E-3,
                     -internal_loss * 1E-3
                     ]
            flows = [round(i) for i in flows]
            flows_labels = ['Heating\nby Residents',
                            'Solar\nGain',
                            'Heater',
                            'GainedEnergy',
                            'Heat\nLoss']

            orientations = [1, 0, -1, -1, 1]
        pathlengths = 0.4

        fig = plt.figure(figsize=[10, 10])
        axis = fig.add_subplot(1, 1, 1, xticks=[], yticks=[])

        sankey = mpl.sankey.Sankey(ax=axis, scale=5E-5, offset=3E-1, head_angle=100, margin=0.4)
        sankey.add(flows=flows,
                   labels=flows_labels,
                   orientations=orientations,
                   pathlengths=pathlengths)
        sankey.finish()
        plt.title("Residence Annual Thermal Equilibrium [kWh]", fontsize=18)
        plt.axis("off")
        plt.savefig(self.filepath)
        plt.close()

        flows = [heating_by_residents * 1E-3,
                 solar_gain_through_windows * 1E-3,
                 thermal_energy_delivered * 1E-3,
                 stored_energy_variation * 1E-3,
                 internal_loss * 1E-3,
                 total_energy_to_residence * 1E-3
                 ]

    def calculate_metrics(self, data):
        """ Calculates the metrics for the sankey chart. """
        heating_by_residents = 0
        total_energy_to_residence = 0
        solar_gain_through_windows = 0
        internal_loss = 0
        stored_energy_variation = 0
        thermal_energy_delivered = 0
        for _index, output_result in enumerate(data):
            if output_result.component_name == "Occupancy" and \
                    output_result.display_name == loadprofilegenerator_connector.Occupancy.HeatingByResidents:
                heating_by_residents = sum(output_result.Results)
            if output_result.component_name == "HeatPump" and (output_result.display_name in
                                                               (generic_heat_pump_modular.ModularHeatPump.ThermalPowerDelivered,
                                                                generic_heat_pump.GenericHeatPump.ThermalEnergyDelivered,
                                                                advanced_heat_pump_hplib.HeatPumpHplib.ThermalOutputPower)):
                thermal_energy_delivered = sum(output_result.Results)
            if output_result.component_name == "Building":
                if output_result.display_name == building.Building.TotalEnergyToResidence:
                    total_energy_to_residence = sum(output_result.Results)
                if output_result.display_name == building.Building.SolarGainThroughWindows:
                    solar_gain_through_windows = sum(output_result.Results)
                if output_result.display_name == building.Building.InternalLoss:
                    internal_loss = sum(output_result.Results)
                if output_result.display_name == building.Building.StoredEnergyVariation:
                    stored_energy_variation = sum(output_result.Results)
        if heating_by_residents == 0 or total_energy_to_residence == 0 or solar_gain_through_windows == 0:
            raise Exception("Sum of outputs has not been calculated.")
        return heating_by_residents, internal_loss, solar_gain_through_windows, \
            stored_energy_variation, thermal_energy_delivered, total_energy_to_residence



"""LoadProfile Generator Connector Module."""

# Generic/Built-in
import json
from typing import Any
from os import path, listdir, makedirs
import shutil
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import pandas as pd
import numpy as np

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters
from hisim.components.configuration import HouseholdWarmWaterDemandConfig
from hisim.components.configuration import PhysicsConfig

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass
class OccupancyConfig(cp.ConfigBase):
    def get_main_classname(self):
        return Occupancy.get_full_classname()

    name: str
    profile_name: str

    @classmethod
    def get_default_CHS01(cls) -> Any:
        config = OccupancyConfig("Occupancy_1", "CH01")
        return config


class Occupancy(cp.Component):
    """
    Class component that provides heating generated, the electricity consumed
    by the residents. Data provided or based on LPG exports.

    Parameters
    -----------------------------------------------
    profile: string
        profile code corresponded to the family or residents configuration

    ComponentInputs:
    -----------------------------------------------
       None

    ComponentOutputs:
    -----------------------------------------------
       Number of Residents: Any
       Heating by Residents: W
       Electricity Consumption: kWh
       Water Consumption: L
    """

    # Inputs
    WW_MassInput = "Warm Water Mass Input"  # kg/s
    WW_TemperatureInput = "Warm Water Temperature Input"  # °C

    # Outputs
    # output
    WW_MassOutput = "Mass Output"  # kg/s
    WW_TemperatureOutput = "Temperature Output"  # °C
    EnergyDischarged = "Energy Discharged"  # W
    DemandSatisfied = "Demand Satisfied"  # 0 or 1

    NumberByResidents = "NumberByResidents"
    HeatingByResidents = "HeatingByResidents"
    ElectricityOutput = "ElectricityOutput"
    WaterConsumption = "WaterConsumption"

    Electricity_Demand_Forecast_24h = "Electricity_Demand_Forecast_24h"

    # Similar components to connect to:
    # None
    @utils.measure_execution_time
    def __init__(
        self, my_simulation_parameters: SimulationParameters, config: OccupancyConfig
    ) -> None:
        super().__init__(
            name="Occupancy", my_simulation_parameters=my_simulation_parameters
        )
        self.profile_name = config.profile_name
        self.occupancyConfig = config
        self.build()

        # Inputs - Not Mandatories
        self.ww_mass_input: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_MassInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.KG_PER_SEC,
            False,
        )
        self.ww_temperature_input: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_TemperatureInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            False,
        )

        # Outputs
        # self.ww_mass_output: cp.ComponentOutput = self.add_output(self.ComponentName,
        #                                                       self.WW_MassOutput,
        #                                                       lt.LoadTypes.WarmWater, lt.Units.kg_per_sec)
        # self.ww_temperature_output: cp.ComponentOutput = self.add_output(self.ComponentName,
        #                                                              self.WW_TemperatureOutput,
        #                                                              lt.LoadTypes.WarmWater,
        #                                                              lt.Units.Celsius)

        # self.energy_discharged: cp.ComponentOutput = self.add_output(self.ComponentName, self.EnergyDischarged, lt.LoadTypes.WarmWater, lt.Units.Watt)
        # self.demand_satisfied: cp.ComponentOutput = self.add_output(self.ComponentName, self.DemandSatisfied, lt.LoadTypes.WarmWater, lt.Units.Any)

        self.number_of_residentsC: cp.ComponentOutput = self.add_output(
            self.component_name, self.NumberByResidents, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.heating_by_residentsC: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.HeatingByResidents,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
        )
        self.electricity_outputC: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED
            ],
        )

        self.water_consumptionC: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.WaterConsumption,
            lt.LoadTypes.WARM_WATER,
            lt.Units.LITER,
        )

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(
        self, timestep: int, stsv: cp.SingleTimeStepValues, force_conversion: bool
    ) -> None:
        if self.ww_mass_input.source_output is not None:
            # ww demand
            ww_temperature_demand = HouseholdWarmWaterDemandConfig.ww_temperature_demand

            # From Thermal Energy Storage
            ww_mass_input_per_sec = stsv.get_input_value(self.ww_mass_input)  # kg/s
            # ww_mass_input = ww_mass_input_per_sec * self.seconds_per_timestep           # kg
            ww_mass_input: float = ww_mass_input_per_sec
            ww_temperature_input = stsv.get_input_value(self.ww_temperature_input)  # °C

            # Information import
            freshwater_temperature = (
                HouseholdWarmWaterDemandConfig.freshwater_temperature
            )
            temperature_difference_hot = (
                HouseholdWarmWaterDemandConfig.temperature_difference_hot
            )  # Grädigkeit
            temperature_difference_cold = (
                HouseholdWarmWaterDemandConfig.temperature_difference_cold
            )
            energy_losses_watt = HouseholdWarmWaterDemandConfig.heat_exchanger_losses
            # energy_losses = energy_losses_watt * self.seconds_per_timestep
            energy_losses = 0
            specific_heat = 4180 / 3600

            ww_energy_demand = (
                specific_heat
                * self.water_consumption[timestep]
                * (ww_temperature_demand - freshwater_temperature)
            )

            if (
                ww_temperature_input
                > (ww_temperature_demand + temperature_difference_hot)
                or ww_energy_demand == 0
            ):
                demand_satisfied = 1
            else:
                demand_satisfied = 0

            if ww_energy_demand > 0 and (
                ww_mass_input == 0 and ww_temperature_input == 0
            ):
                """first iteration --> random numbers"""
                ww_temperature_input = 40.45
                ww_mass_input = 9.3

            """
            Warm water is provided by the warmwater stoage.
            The household needs water at a certain temperature. To get the correct temperature the amount of water from
            the wws is regulated and is depending on the temperature provided by the wws. The backflowing water to wws
            is cooled down to the temperature of (freshwater+temperature_difference_cold) --> ww_temperature_output.
            """
            if ww_energy_demand > 0:
                # heating up the freshwater. The mass is consistent
                energy_discharged = ww_energy_demand + energy_losses
                ww_temperature_output: float = (
                    freshwater_temperature + temperature_difference_cold
                )
                ww_mass_input = energy_discharged / (
                    PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
                    * (ww_temperature_input - ww_temperature_output)
                )
            else:
                ww_temperature_output = ww_temperature_input
                ww_mass_input = 0
                energy_discharged = 0

            ww_mass_output = ww_mass_input

            # stsv.set_output_value(self.ww_mass_output, ww_mass_output)
            # stsv.set_output_value(self.ww_temperature_output, ww_temperature_output)
            # stsv.set_output_value(self.demand_satisfied, demand_satisfied)
            # stsv.set_output_value(self.energy_discharged, energy_discharged)

        stsv.set_output_value(
            self.number_of_residentsC, self.number_of_residents[timestep]
        )
        stsv.set_output_value(
            self.heating_by_residentsC, self.heating_by_residents[timestep]
        )
        stsv.set_output_value(
            self.electricity_outputC, self.electricity_consumption[timestep]
        )
        stsv.set_output_value(self.water_consumptionC, self.water_consumption[timestep])

        if self.my_simulation_parameters.system_config.predictive == True:
            last_forecast_timestep = int( timestep + 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep )
            if (last_forecast_timestep > len(self.electricity_consumption)):
                last_forecast_timestep = len(self.electricity_consumption)
            #log.information( type(self.temperature))
            demandforecast = self.electricity_consumption[ timestep : last_forecast_timestep ]
            self.simulation_repository.set_entry( self.Electricity_Demand_Forecast_24h, demandforecast )

        if self.my_simulation_parameters.system_config.predictive == True:
            last_forecast_timestep = int(
                timestep
                + 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            )
            if last_forecast_timestep > len(self.electricity_consumption):
                last_forecast_timestep = len(self.electricity_consumption)
            # log.information( type(self.temperature))
            demandforecast = self.electricity_consumption[
                timestep:last_forecast_timestep
            ]
            self.simulation_repository.set_entry(
                self.Electricity_Demand_Forecast_24h, demandforecast
            )

    def build(self):
        file_exists, cache_filepath = utils.get_cache_file(
            component_key=self.component_name,
            parameter_class=self.occupancyConfig,
            my_simulation_parameters=self.my_simulation_parameters,
        )

        # create directories to put in files for cars and smart devices
        for tag in ['utsp_reports', 'utsp_results']:
            isExist = path.exists(utils.HISIMPATH[tag])
            if not isExist:
               # Create a new directory because it does not exist
               makedirs(utils.HISIMPATH[tag])

        # copy example files for ev and smart devices to right folder
        reportfiles = listdir(utils.HISIMPATH['utsp_example_reports'])
        for file in reportfiles:
            shutil.copyfile(path.join(utils.HISIMPATH['utsp_example_reports'], file), path.join(utils.HISIMPATH['utsp_reports'], file))
        resultfiles = listdir(utils.HISIMPATH['utsp_example_results'])
        for file in resultfiles:
            shutil.copyfile(path.join(utils.HISIMPATH['utsp_example_results'], file), path.join(utils.HISIMPATH['utsp_results'], file))

        if file_exists:
            dataframe = pd.read_csv(
                cache_filepath, sep=",", decimal=".", encoding="cp1252"
            )
            self.number_of_residents = dataframe["number_of_residents"].tolist()
            self.heating_by_residents = dataframe["heating_by_residents"].tolist()
            self.electricity_consumption = dataframe["electricity_consumption"].tolist()
            self.water_consumption = dataframe["water_consumption"].tolist()
        else:
            ################################
            # Calculates heating generated by residents and loads number of residents
            # Heat power generated per resident in W
            # mode 1: awake
            # mode 2: sleeping
            gain_per_person = [150, 100]

            # load occupancy profile
            occupancy_profile = []
            filepaths = utils.HISIMPATH["occupancy"][self.profile_name][
                "number_of_residents"
            ]
            for filepath in filepaths:
                with open(filepath) as json_file:
                    json_filex = json.load(json_file)
                occupancy_profile.append(json_filex)

            # see how long csv files from LPG are to check if averaging has to be done and calculate desired length
            steps_original = len(occupancy_profile[0]["Values"])
            steps_desired = int(
                365 * 24 * (3600 / self.my_simulation_parameters.seconds_per_timestep)
            )
            steps_ratio = int(steps_original / steps_desired)

            # initialize number of residence and heating by residents
            self.heating_by_residents = [0] * steps_desired
            self.number_of_residents = [0] * steps_desired

            # load electricity consumption and water consumption
            pre_electricity_consumption = pd.read_csv(
                utils.HISIMPATH["occupancy"][self.profile_name][
                    "electricity_consumption"
                ],
                sep=";",
                decimal=",",
                encoding="cp1252",
            )
            pre_water_consumption = pd.read_csv(
                utils.HISIMPATH["occupancy"][self.profile_name]["water_consumption"],
                sep=";",
                decimal=",",
                encoding="cp1252",
            )

            # convert electricity consumption and water consumption to desired format and unit
            self.electricity_consumption = pd.to_numeric(
                pre_electricity_consumption["Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60W / min
            self.water_consumption = pd.to_numeric(
                pre_water_consumption["Sum [L]"]
            ).tolist()

            # process data when time resolution of inputs matches timeresolution of simulation
            if steps_original == steps_desired:
                for mode in range(len(gain_per_person)):
                    for timestep in range(steps_original):
                        self.number_of_residents[timestep] += occupancy_profile[mode][
                            "Values"
                        ][timestep]
                        self.heating_by_residents[timestep] = (
                            self.heating_by_residents[timestep]
                            + gain_per_person[mode]
                            * occupancy_profile[mode]["Values"][timestep]
                        )

            # average data, when time resolution of inputs is coarser than time resolution of simulation
            elif steps_original > steps_desired:
                for mode in range(len(gain_per_person)):
                    for timestep in range(steps_desired):
                        number_of_residents_av = (
                            sum(
                                occupancy_profile[mode]["Values"][
                                    timestep
                                    * steps_ratio:(timestep + 1)
                                    * steps_ratio
                                ]
                            )
                            / steps_ratio
                        )
                        self.number_of_residents[timestep] += np.round(
                            number_of_residents_av
                        )
                        self.heating_by_residents[timestep] = (
                            self.heating_by_residents[timestep]
                            + gain_per_person[mode] * number_of_residents_av
                        )
                # power needs averaging, not sum
                self.electricity_consumption = [
                    sum(self.electricity_consumption[n:n + steps_ratio]) / steps_ratio
                    for n in range(0, steps_original, steps_ratio)
                ]
                self.water_consumption = [
                    sum(self.water_consumption[n:n + steps_ratio])
                    for n in range(0, steps_original, steps_ratio)
                ]

            else:
                raise Exception(
                    "input from LPG is given in wrong time resolution - or at least cannot be interpolated correctly"
                )

            # Saves data in cache
            data = np.transpose(
                [
                    self.number_of_residents,
                    self.heating_by_residents,
                    self.electricity_consumption,
                    self.water_consumption,
                ]
            )
            database = pd.DataFrame(
                data,
                columns=[
                    "number_of_residents",
                    "heating_by_residents",
                    "electricity_consumption",
                    "water_consumption",
                ],
            )

            database.to_csv(cache_filepath)
            del data
            del database
            # utils.save_cache("Occupancy", parameters, database)
        self.max_hot_water_demand = max(self.water_consumption)

    def write_to_report(self):
        """Writes a report."""
        lines = []
        lines.append(f"Name: {self.component_name}")
        lines.append(f"Profile: {self.profile_name}")
        return lines



from typing import Any
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim import loadtypes as lt

from hisim.components.configuration import CHPControllerConfig, GasControllerConfig, AdvElectrolyzerConfig
from hisim.components import advanced_fuel_cell as chp
from hisim.components.configuration import ExtendedControllerConfig
from hisim.simulationparameters import SimulationParameters
from hisim import log
from math import ceil
from copy import deepcopy

class ExtendedControllerSimulation:
    def __init__(self, config: ExtendedControllerConfig):
        pass

    def regulate_chp_mode_power(self, state_chp :Any, runtime_chp: Any, power_supply_pv: Any, electricity_demand_household: Any, seconds_per_timestep: Any) -> Any:
        """
        :param power_supply_pv:
        :param electricity_demand_household:
        :param seconds_per_timestep:
        :return:
        """
        demand = electricity_demand_household - power_supply_pv
        if chp.CHPConfig.is_modulating:
            power_states_possible = ExtendedControllerConfig.chp_power_states_possible
            min_power = chp.CHPConfig.P_el_min
            max_power = chp.CHPConfig.P_el_max
            power_delta = max_power - min_power
            # the first step is already the minimum power
            power_per_step_size = power_delta / (power_states_possible - 1)

            if demand <= 0:
                if state_chp == 0:
                    pass
                else:
                    minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                    if runtime_chp >= minimum_timesteps:
                        # switch off chp
                        state_chp = 0
                        runtime_chp = 0
                    else:
                        state_chp = 1 / power_states_possible
            elif 0 < demand < chp.CHPConfig.P_el_min:
                # maximum_autarky --> production time if P > 0 or production only if P in range?
                if ExtendedControllerConfig.maximum_autarky:
                    state_chp = 1 / power_states_possible  # eg 0.1
                else:
                    if state_chp == 0:
                        pass  # state_chp = 0
                    else:
                        minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                        if runtime_chp >= minimum_timesteps:
                            # switch off chp
                            state_chp = 0
                            runtime_chp = 0
                        else:
                            state_chp = 1 / power_states_possible
            elif chp.CHPConfig.P_el_min <= demand < chp.CHPConfig.P_el_max:
                # minimum power plus the demand depending power
                state_chp = (1 + ceil((demand - chp.CHPConfig.P_el_min) / power_per_step_size)) / power_states_possible
                assert (1 / power_states_possible) < state_chp <= 1
            else:  # demand >= chp.CHPConfig.P_el_max
                state_chp = 1

            if state_chp == 0:
                generated_electricity = 0
                power_from_or_to_grid = demand - generated_electricity
            elif 0 < state_chp <= 1:
                generated_electricity = chp.CHPConfig.P_el_min + power_per_step_size * (state_chp * power_states_possible - 1)
                power_from_or_to_grid = demand - generated_electricity
            else:
                log.error("Wrong controller state")
                raise ValueError

        else:
            if demand <= 0:
                if state_chp == 0:
                    power_from_or_to_grid = demand
                else:
                    minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                    if runtime_chp >= minimum_timesteps:
                        state_chp = 0
                        power_from_or_to_grid = demand
                    else:
                        state_chp = 1
                        power_from_or_to_grid = demand - chp.CHPConfig.P_el_max

            else:  # demand > 0
                state_chp = 1
                power_from_or_to_grid = demand - chp.CHPConfig.P_el_max

        # independent of modulation
        if state_chp > 0:
            runtime_chp += 1
        else:
            runtime_chp = 0

        return state_chp, runtime_chp, power_from_or_to_grid

    def regulate_chp_mode_heat(self, temperatures_in_tank, previous_state, runtime_chp, pv_production, electricity_demand_household, seconds_per_timestep):
        # the heat model has no modulation because there is a buffer (warm water storage)

        heights_in_tank = CHPControllerConfig.heights_in_tank

        if (CHPControllerConfig.height_upper_sensor or CHPControllerConfig.height_lower_sensor) not in heights_in_tank:
            log.error("Wrong sensor setting. Only 0, 20, 40, 60, 80, 100% are allowed.\n"
                  "You tried " + str(CHPControllerConfig.height_upper_sensor) + " and " + str(CHPControllerConfig.height_lower_sensor))
            raise ValueError

        # get temperatures at the chosen sensors
        for i in range(len(heights_in_tank)):
            if CHPControllerConfig.height_upper_sensor == heights_in_tank[i]:
                temperature_upper_sensor = temperatures_in_tank[i]
            if CHPControllerConfig.height_lower_sensor == heights_in_tank[i]:
                temperature_lower_sensor = temperatures_in_tank[i]
        state_chp = previous_state
        # upper sensor
        if temperature_upper_sensor < CHPControllerConfig.temperature_switch_on:
            # switch on
            state_chp = 1

        # lower sensor (no check needed if chp is off)
        if state_chp == 1:
            if temperature_lower_sensor > CHPControllerConfig.temperature_switch_off:
                minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                if runtime_chp > minimum_timesteps:
                    # switch off is possible if chp has run at least xx min
                    state_chp = 0

        # increase counter
        if state_chp == 1:
            runtime_chp += 1
        elif state_chp == 0:
            runtime_chp = 0
        else:
            log.error("Wrong state_chp")
            raise ValueError
        # 'easy equation' --> no modulation
        power_from_or_to_grid = electricity_demand_household - pv_production - chp.CHPConfig.P_el_max * state_chp

        return state_chp, runtime_chp, power_from_or_to_grid

    def regulate_gas_heater(self, temperatures_in_tank: Any, previous_state: float, runtime_counter: int, seconds_per_timestep: int) -> Any:
        # the gas_heater model has no modulation because there is a buffer (warm water storage)
        # ToDo: future --> make modulating possible if the waste energy is to high? reduce power, increase mass_flow

        heights_in_tank = CHPControllerConfig.heights_in_tank
        if (GasControllerConfig.height_upper_sensor or GasControllerConfig.height_lower_sensor) not in heights_in_tank:
            log.error("Wrong sensor setting. Only 0, 20, 40, 60, 80, 100% are allowed.\n"
                  "You tried " + str(GasControllerConfig.height_upper_sensor) + " and " + str(GasControllerConfig.height_lower_sensor))
            raise ValueError

        for i in range(len(heights_in_tank)):
            if GasControllerConfig.height_upper_sensor == heights_in_tank[i]:
                temperature_upper_sensor = temperatures_in_tank[i]
            if GasControllerConfig.height_lower_sensor == heights_in_tank[i]:
                temperature_lower_sensor = temperatures_in_tank[i]
        state_gas_heater = previous_state
        # upper sensor
        if temperature_upper_sensor < GasControllerConfig.temperature_switch_on:
            # switch on
            state_gas_heater = 1

        # lower sensor (no check needed if gas heater is off)
        if state_gas_heater == 1:
            if temperature_lower_sensor > GasControllerConfig.temperature_switch_off:
                minimum_timesteps = (GasControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                if runtime_counter > minimum_timesteps:
                    # switch off is possible if gas_heater has run at least xx min
                    state_gas_heater = 0
            else:
                state_gas_heater = 1

        # increase counter
        if state_gas_heater == 1:
            runtime_counter += 1
        elif state_gas_heater == 0:
            runtime_counter = 0
        else:
            log.error("Wrong state_chp")
            raise ValueError

        return state_gas_heater, runtime_counter

    def power_distribution_to_electrolyzer(self, power_from_or_to_grid: Any) -> Any:
        """
        Hydrogen storage must be able to store the produced massflow of hydrogen.
        Otherwise the dimensioning of the system is incorrect --> is checked in hydrogen_storage
        power_from_or_to_grid is positive if there is a demand of energy and negative if there is a surplus of energy.
        power_available is defined the other way round --> negate the value
        """
        power_available = -power_from_or_to_grid

        if power_available < AdvElectrolyzerConfig.min_power:
            power_to_electrolyzer = 0
            # no change
            # power_from_or_to_grid = power_available
        elif AdvElectrolyzerConfig.min_power <= power_available <= AdvElectrolyzerConfig.max_power:
            power_to_electrolyzer = power_available
            power_from_or_to_grid = 0
        else: # power_available > ElectrolyzerConfig.max_power:
            power_to_electrolyzer = AdvElectrolyzerConfig.max_power
            # not al the electricity can go to the electolyzer --> power_from_or_to_gridstays negative
            power_from_or_to_grid = power_to_electrolyzer - power_available

        return power_to_electrolyzer, power_from_or_to_grid

class ExtendedController(Component):
    # inputs
    ElectricityDemand = "Electricity Demand"  # W
    PV_Production = "PV Production"  # W

    # temperatures (input)
    Temperature0Percent = "Temperature 0 Percent"  # °C
    Temperature20Percent = "Temperature 20 Percent"  # °C
    Temperature40Percent = "Temperature 40 Percent"  # °C
    Temperature60Percent = "Temperature 60 Percent"  # °C
    Temperature80Percent = "Temperature 80 Percent"  # °C
    Temperature100Percent = "Temperature 100 Percent"  # °C

    # Output
    ControllerCHP = "Controller CHP"
    ControllerGasHeater = "Controller Gas Heater"
    PowerToElectrolyzer = "Power To Electrolyzer"
    PowerFromOrToGrid = "Power From Or To Grid"

    RuntimeCounterCHP = "RuntimeCounterCHP"
    RuntimeCounterGasHeater = "RuntimeCounterGasHeater"

    def __init__(self, component_name:str, config: ExtendedControllerConfig, my_simulation_parameters: SimulationParameters ) -> None:
        super().__init__(name=component_name,  my_simulation_parameters=my_simulation_parameters)
        # Input
        self.electricity_demand_household: ComponentInput = self.add_input(self.component_name, ExtendedController.ElectricityDemand, lt.LoadTypes.ELECTRICITY, lt.Units.WATT, True)
        self.pv_production: ComponentInput = self.add_input(self.component_name, ExtendedController.PV_Production, lt.LoadTypes.ELECTRICITY, lt.Units.WATT, True)

        self.temperature_0_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature0Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.temperature_20_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature20Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.temperature_40_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature40Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.temperature_60_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature60Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.temperature_80_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature80Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.temperature_100_percent: ComponentInput = self.add_input(self.component_name, ExtendedController.Temperature100Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)

        # Output
        self.controller_chp: ComponentOutput = self.add_output(self.component_name, ExtendedController.ControllerCHP, lt.LoadTypes.ANY, lt.Units.PERCENT)
        self.controller_gas_heater: ComponentOutput = self.add_output(self.component_name, ExtendedController.ControllerGasHeater, lt.LoadTypes.ANY, lt.Units.PERCENT)
        self.power_to_electrolyzer: ComponentOutput = self.add_output(self.component_name, ExtendedController.PowerToElectrolyzer, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)
        self.power_from_or_to_grid: ComponentOutput = self.add_output(self.component_name, ExtendedController.PowerFromOrToGrid, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)

        self.runtime_counter_chp: ComponentOutput = self.add_output(self.component_name, ExtendedController.RuntimeCounterCHP, lt.LoadTypes.ANY, lt.Units.ANY)
        self.runtime_counter_gas_heater: ComponentOutput = self.add_output(self.component_name, ExtendedController.RuntimeCounterGasHeater, lt.LoadTypes.ANY, lt.Units.ANY)

        self.extended_controller = ExtendedControllerSimulation(config)
        self.seconds_per_timestep = my_simulation_parameters.seconds_per_timestep

        # CHP state/runtime & Gas state/runtime
        self.state_chp1 = 0
        self.runtime_chp1 = 0
        self.state_gas_heater1 = 0
        self.runtime_gas_heater1 = 0

        # self.previous_state = [self.state_chp1, self.runtime_chp1, self.state_gas_heater1, self.runtime_gas_heater1]
        self.previous_state_chp1 = self.state_chp1
        self.previous_runtime_chp1 = self.runtime_chp1
        self.previous_state_gas_heater1 = self.state_gas_heater1
        self.previous_runtime_gas_heater1 = self.runtime_gas_heater1

        self.test_pv: float = 0
        self.test_grid: float = 0
        self.test_electrolyzer: float = 0
        self.test_demand: float = 0
        self.test_state:float = 0

    def i_save_state(self) -> None:
        # self.previous_state = self.extended_controller.begin_new_timestep()
        self.previous_state_chp1 = deepcopy(self.state_chp1)
        self.previous_runtime_chp1 = deepcopy(self.runtime_chp1)
        self.previous_state_gas_heater1 = deepcopy(self.state_gas_heater1)
        self.previous_runtime_gas_heater1 = deepcopy(self.runtime_gas_heater1)

    def i_restore_state(self) -> None:
        # self.extended_controller.reset_to_last_timestep(self.previous_state)
        self.state_chp1 = deepcopy(self.previous_state_chp1)
        self.runtime_chp1 = deepcopy(self.previous_runtime_chp1)
        self.state_gas_heater1 = deepcopy(self.previous_state_gas_heater1)
        self.runtime_gas_heater1 = deepcopy(self.previous_runtime_gas_heater1)

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        if force_convergence:
            return

        # Inputs
        electricity_demand_household = stsv.get_input_value(self.electricity_demand_household)
        pv_production: float = stsv.get_input_value(self.pv_production)

        # not needed for power
        if ExtendedControllerConfig.chp_mode == "heat" or ExtendedControllerConfig.gas_heater:
            temperature_0_percent = stsv.get_input_value(self.temperature_0_percent)
            temperature_20_percent = stsv.get_input_value(self.temperature_20_percent)
            temperature_40_percent = stsv.get_input_value(self.temperature_40_percent)
            temperature_60_percent = stsv.get_input_value(self.temperature_60_percent)
            temperature_80_percent = stsv.get_input_value(self.temperature_80_percent)
            temperature_100_percent = stsv.get_input_value(self.temperature_100_percent)

            temperatures_in_tank = [temperature_0_percent, temperature_20_percent, temperature_40_percent,
                                    temperature_60_percent, temperature_80_percent, temperature_100_percent]

        # Combined heat and power plant
        if ExtendedControllerConfig.chp:
            if ExtendedControllerConfig.chp_mode == "power":
                # chp is running depending on the delta between pv_production and electricity_demand_household
                self.state_chp1, self.runtime_chp1, power_from_or_to_grid = self.extended_controller.regulate_chp_mode_power(self.state_chp1,
                                                                                                                             self.runtime_chp1,
                                                                                                                             pv_production,
                                                                                                                             electricity_demand_household,
                                                                                                                             self.seconds_per_timestep)
            elif ExtendedControllerConfig.chp_mode == "heat":
                self.state_chp1, self.runtime_chp1, power_from_or_to_grid = self.extended_controller.regulate_chp_mode_heat(temperatures_in_tank, self.state_chp1, self.runtime_chp1, pv_production, electricity_demand_household, self.seconds_per_timestep)
            else:
                log.error("Wrong chp controller settings! Choose between heat and power")
                raise ValueError
        else:
            power_from_or_to_grid = pv_production - electricity_demand_household

        # Gas heater
        if ExtendedControllerConfig.gas_heater:
            self.state_gas_heater1, self.runtime_gas_heater1 = self.extended_controller.regulate_gas_heater(temperatures_in_tank, self.state_gas_heater1, self.runtime_gas_heater1, self.seconds_per_timestep)

        if not ExtendedControllerConfig.chp or not ExtendedControllerConfig.gas_heater:
            log.error("Choose a energy source")
            raise ValueError

        # Electrolyzer
        # ToDo: Sollte er laufen wenn die KWK auch läuft... (stromgeführt nein, wärmegefüht evtl ja slange keine Batterie da ist)

        if ExtendedControllerConfig.electrolyzer:
            power_to_electrolyzer, power_from_or_to_grid = self.extended_controller.power_distribution_to_electrolyzer(power_from_or_to_grid)
        else:
            power_to_electrolyzer = 0

        # Outputs
        stsv.set_output_value(self.controller_chp, self.state_chp1)
        # stsv.set_output_value(self.controller_gas_heater, self.extended_controller.state_gas_heater)
        stsv.set_output_value(self.controller_gas_heater, self.state_gas_heater1)

        stsv.set_output_value(self.power_to_electrolyzer, power_to_electrolyzer)
        stsv.set_output_value(self.power_from_or_to_grid, power_from_or_to_grid)

        stsv.set_output_value(self.runtime_counter_chp, self.runtime_chp1)
        stsv.set_output_value(self.runtime_counter_gas_heater, self.runtime_gas_heater1)


        # test
        self.test_demand = electricity_demand_household
        self.test_pv = pv_production
        self.test_grid = power_from_or_to_grid
        self.test_electrolyzer = power_to_electrolyzer
        self.test_state = self.state_chp1

        # self.test_state = self.extended_controller.state_chp

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        # check the electricity balance
        if chp.CHPConfig.is_modulating:
            if self.test_state > 0:
                power_chp_test = chp.CHPConfig.P_el_min + (chp.CHPConfig.P_el_max - chp.CHPConfig.P_el_min) / (ExtendedControllerConfig.chp_power_states_possible - 1) * (self.test_state * ExtendedControllerConfig.chp_power_states_possible - 1)
            else:
                power_chp_test = 0

            if 0.00001 > (self.test_pv + power_chp_test + self.test_grid - self.test_demand - self.test_electrolyzer):
                pass
            else:
                log.error("Wrong energy balance:")
                log.error("State CHP: " + str(self.test_state))
                log.error("test_pv: " + str(self.test_pv))
                log.error("power_chp_test: " + str(power_chp_test))
                log.error("test_grid: " + str(self.test_grid))
                log.error("test_demand: " + str(self.test_demand))
                log.error("test_electrolyzer: " + str(self.test_electrolyzer))
                raise ValueError

        else:
            if 0.00001 > (self.test_pv + chp.CHPConfig.P_el_max * self.test_state + self.test_grid - self.test_electrolyzer - self.test_demand):
                pass
            else:
                log.error("Wrong energy balance:")
                log.error("State CHP: " + str(self.test_state))
                log.error("test_pv: " + str(self.test_pv))
                log.error("power_chp_test: " + str(chp.CHPConfig.P_el_max))
                log.error("test_grid: " + str(self.test_grid))
                log.error("test_demand: " + str(self.test_demand))
                log.error("test_electrolyzer: " + str(self.test_electrolyzer))
                raise ValueError



""" Module for generating reports. """
# clean
import copy
import time
import os
from typing import Any
from reportlab.lib.enums import TA_JUSTIFY
from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
from reportlab.platypus import Table

from hisim import utils


class ReportGenerator:

    """ Class for generating reports. """

    def __init__(self, dirpath: str) -> None:
        """ Initializes the pdf report. """
        if dirpath is None:
            raise ValueError("Result path for the report was none.")
        self.story: Any
        self.filepath = os.path.join(dirpath, "report.pdf")
        self.open()
        self.write_preamble()
        self.close()

    def open(self):
        """ Opens a file. """
        self.doc = SimpleDocTemplate(self.filepath, pagesize=letter,
                                     rightMargin=72, leftMargin=72,
                                     topMargin=72, bottomMargin=18)
        self.styles = getSampleStyleSheet()
        self.styles.add(ParagraphStyle(name='Justify', alignment=TA_JUSTIFY))

    def write_preamble(self):
        """ Writes the preamble. """
        # Configuration taken mostly from following tutorial
        # https://www.blog.pythonlibrary.org/2010/03/08/a-simple-step-by-step-reportlab-tutorial/
        story = []

        # Inserts HiSim logo
        logo = os.path.join(utils.hisim_postprocessing_img, "hisim_logo.png")
        im1 = Image(logo, 2 * inch, inch)
        im1.hAlign = "LEFT"

        # Inserts FZJ logo
        logo = os.path.join(utils.hisim_postprocessing_img, "fzj_logo.jpg")
        im2 = Image(logo, 2 * inch, inch)
        im2.hAlign = "RIGHT"

        data = [[im1, im2]]
        report_table = Table(data)
        story.append(report_table)
        story.append(Spacer(1, 24))

        # Inserts authors
        authors = ["Developers: Dr. Noah Pflugradt",
                   "Vitor Hugo Bellotto Zago"]
        for part in authors:
            ptext = f'<font size="12">{part.strip()}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
        story.append(Spacer(1, 12))

        # Inserts address
        address_parts = ["Forschungszentrum Jülich",
                         "Institute of Energy and Climate Research",
                         "Techno - Economic Systems Analysis (IEK - 3)",
                         "Wilhelm - Johnen - Straße",
                         "52428 Jülich",
                         "Germany"]
        for part in address_parts:
            ptext = f'<font size="12">{part.strip()}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
        story.append(Spacer(1, 12))

        # Inserts time
        formatted_time = time.ctime()
        ptext = f'<font size="12">{formatted_time}</font>'
        story.append(Paragraph(ptext, self.styles["Normal"]))
        story.append(Spacer(1, 12))

        if hasattr(self, "executation_time"):
            # formatted_time
            ptext = f'<font size="12">{formatted_time}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
            story.append(Spacer(1, 12))

        self.story = story

    def write(self, text):
        """ Writes a paragraph. """
        if len(text) != 0:
            bar_string = "=============================================================="
            self.story.append(Paragraph(bar_string, self.styles["Normal"]))
            for part in text:
                ptext = f'<font size="12">{part}</font>'
                self.story.append(Paragraph(ptext, self.styles["Normal"]))
            self.story.append(Spacer(1, 12))

    def get_story(self):
        """ Gets the story. """
        self.story = copy.deepcopy(self.story)

    def close(self):
        """ Closes the report. """
        story = copy.deepcopy(self.story)
        self.doc.build(story)



""" Contains functions for initializing and connecting components.

The functions are all called in modular_household.
"""

from typing import List, Optional, Tuple, Any
from os import listdir, path
import json

from utspclient.helpers.lpgpythonbindings import JsonReference
from building_sizer.heating_system_enums import HeatingSystems

import hisim.loadtypes as lt
from hisim.component import Component
from hisim.simulator import SimulationParameters
from hisim.components import generic_heat_pump_modular
from hisim.components import generic_heat_source
from hisim.components import controller_l1_building_heating
from hisim.components import controller_l1_heatpump
from hisim.components import controller_l2_generic_heat_simple
from hisim.components import controller_l2_energy_management_system
from hisim.components import generic_hot_water_storage_modular
from hisim.components import loadprofilegenerator_connector
from hisim.components import weather
from hisim.components import building
from hisim.components import generic_pv_system
from hisim.components import generic_smart_device
from hisim.components import generic_car
from hisim.components import controller_l1_generic_ev_charge
from hisim.components import advanced_battery_bslib
from hisim.components import advanced_ev_battery_bslib
from hisim.components import generic_CHP
from hisim.components import generic_electrolyzer
from hisim.components import generic_hydrogen_storage
from hisim.components.configuration import HouseholdWarmWaterDemandConfig
from hisim import utils


def configure_pv_system(my_sim: Any, my_simulation_parameters: SimulationParameters, my_weather: weather.Weather,
                        production: List, pv_peak_power: Optional[float], count: int) -> Tuple[List, int]:
    """ Sets PV System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_weather: Weather
        The initialized Weather component.
    production: List
        List of Components with Parameter Production.
    pv_peak_power: float or None
        The peak power of the PV panel. In case of None default is used.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if pv_peak_power is not None:
        my_pv_system_config = generic_pv_system.PVSystem.get_default_config(power=pv_peak_power, source_weight=count)
    else:
        my_pv_system_config = generic_pv_system.PVSystem.get_default_config(source_weight=count)
    count += 1
    my_pv_system = generic_pv_system.PVSystem(
        my_simulation_parameters=my_simulation_parameters, my_simulation_repository=my_sim.simulation_repository,
        config=my_pv_system_config)
    my_pv_system.connect_only_predefined_connections(my_weather)
    my_sim.add_component(my_pv_system)
    production.append(my_pv_system)

    return production, count


def configure_smart_devices(my_sim: Any, my_simulation_parameters: SimulationParameters, count: int, smart_devices_included: bool) \
        -> Tuple[List[generic_smart_device.SmartDevice], int]:
    """ Sets smart devices without controllers.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    smart_devices_included: bool
        True if smart devices (washing machine, dish washer, etc.) are actually smart or surplus controlled.
    count: int
        Integer tracking component hierachy for EMS.

    """
    filepath = path.join(utils.HISIMPATH["utsp_reports"], "FlexibilityEvents.HH1.json")
    device_collection = []
    with open(filepath, mode='r', encoding="utf-8") as jsonfile:
        strfile = json.load(jsonfile)

    for elem in strfile:
        if elem['Device']['Name'] in device_collection:
            pass
        else:
            device_collection.append(elem['Device']['Name'])

    # create all smart devices
    my_smart_devices: List[generic_smart_device.SmartDevice] = []
    for device in device_collection:
        my_smart_devices.append(generic_smart_device.SmartDevice(
            identifier=device, source_weight=count, my_simulation_parameters=my_simulation_parameters,
            smart_devices_included=smart_devices_included))
        my_sim.add_component(my_smart_devices[-1])
        count += 1

    return my_smart_devices, count


def configure_cars(my_sim: Any, my_simulation_parameters: SimulationParameters, count: int, ev_included: bool,
                   occupancy_config: Any) -> Tuple[List[generic_car.Car], int]:
    """ Sets smart devices without controllers.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    count: int
        Integer tracking component hierachy for EMS.
    ev_included: bool
        True if Car is electric, False if it is diesel.
    occupancy_config: loadprofilegenerator_connector.OccupancyConfig
        Unique description of load profile generator call (mobility is related!)


    """
    # get names of all available cars
    filepaths = listdir(utils.HISIMPATH["utsp_results"])
    filepaths_location = [elem for elem in filepaths if "CarLocation." in elem]
    names = [elem.partition(',')[0].partition('.')[2] for elem in filepaths_location]

    # decide if they are diesel driven or electricity driven
    if ev_included:
        my_car_config = generic_car.CarConfig.get_default_ev_config()
    else:
        my_car_config = generic_car.CarConfig.get_default_diesel_config()

    # create all cars
    my_cars: List[generic_car.Car] = []
    for car in names:
        my_car_config.name = car
        my_car_config.source_weight = count
        my_cars.append(generic_car.Car(
            my_simulation_parameters=my_simulation_parameters, config=my_car_config,
            occupancy_config=occupancy_config))
        my_sim.add_component(my_cars[-1])
        count += 1

    return my_cars, count


def configure_ev_batteries(my_sim: Any, my_simulation_parameters: SimulationParameters, my_cars: List[generic_car.Car], charging_station_set: Optional[JsonReference],
                           mobility_set: JsonReference, my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                           clever: bool) -> List:
    """ Sets batteries and controllers of electric vehicles.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_cars: List[Car]
        List of initilized cars.
    charging_station_set: ChargingStationSets
        Encoding of the charging station.
    mobility_set: TransportationDeviceSets
        Encoding of the available cars.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    clever: bool
        True if battery of electric vehicle is charged with surplus.

    """
    ev_capacities = []

    if mobility_set.Name is None:
        raise Exception('For EV configuration mobility set is obligatory.')
    mobility_speed = mobility_set.Name.partition('and ')[2].partition(' ')[2].partition(' km/h')[0]
    if mobility_speed == '30':
        car_battery_config = advanced_ev_battery_bslib.CarBatteryConfig.get_default_config(
            e_bat_custom=30, p_inv_custom=5000, name="CarBattery")
        ev_capacities.append(30)
    elif mobility_speed == '60':
        car_battery_config = advanced_ev_battery_bslib.CarBatteryConfig.get_default_config(
            e_bat_custom=50, p_inv_custom=11000, name="CarBattery")
        ev_capacities.append(50)
    if charging_station_set is None:
        raise Exception('For EV configuration charging station set is obligatory.')

    car_battery_controller_config = controller_l1_generic_ev_charge.ChargingStationConfig.get_default_config(
        charging_station_set=charging_station_set)

    if clever:
        car_battery_controller_config.battery_set = 0.4  # lower threshold for soc of car battery in clever case

    for car in my_cars:
        car_battery_config.source_weight = car.source_weight
        car_battery_controller_config.source_weight = car.source_weight
        my_carbattery = advanced_ev_battery_bslib.CarBattery(my_simulation_parameters=my_simulation_parameters, config=car_battery_config)
        my_controller_carbattery = controller_l1_generic_ev_charge.L1Controller(my_simulation_parameters=my_simulation_parameters,
                                                                                config=car_battery_controller_config)
        my_controller_carbattery.connect_only_predefined_connections(car)
        my_controller_carbattery.connect_only_predefined_connections(my_carbattery)
        my_carbattery.connect_only_predefined_connections(my_controller_carbattery)
        my_sim.add_component(my_carbattery)
        my_sim.add_component(my_controller_carbattery)

        if clever:
            my_electricity_controller.add_component_input_and_connect(
                source_component_class=my_carbattery, source_component_output=my_carbattery.AcBatteryPower,
                source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                source_tags=[lt.ComponentType.CAR_BATTERY, lt.InandOutputType.ELECTRICITY_REAL],
                source_weight=my_carbattery.source_weight)

            electricity_target = my_electricity_controller.add_component_output(
                source_output_name=lt.InandOutputType.ELECTRICITY_TARGET, source_tags=[lt.ComponentType.CAR_BATTERY, lt.InandOutputType.ELECTRICITY_TARGET],
                source_weight=my_controller_carbattery.source_weight, source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT)

            my_controller_carbattery.connect_dynamic_input(
                input_fieldname=controller_l1_generic_ev_charge.L1Controller.ElectricityTarget, src_object=electricity_target)

    return ev_capacities


def configure_smart_controller_for_smart_devices(my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                                                 my_smart_devices: List[generic_smart_device.SmartDevice]) -> None:
    """ Sets l3 controller for smart devices.

    Parameters
    ----------
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_smart_devices: List[SmartDevice]
        List of initilized smart devices.

    """

    for elem in my_smart_devices:
        my_electricity_controller.add_component_input_and_connect(
            source_component_class=elem, source_component_output=elem.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
            source_tags=[lt.ComponentType.SMART_DEVICE, lt.InandOutputType.ELECTRICITY_REAL],
            source_weight=elem.source_weight)

        electricity_to_smart_device = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[lt.ComponentType.SMART_DEVICE, lt.InandOutputType.ELECTRICITY_TARGET],
            source_weight=elem.source_weight, source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT)

        elem.connect_dynamic_input(input_fieldname=generic_smart_device.SmartDevice.ElectricityTarget,
                                   src_object=electricity_to_smart_device)


def configure_battery(my_sim: Any, my_simulation_parameters: SimulationParameters,
                      my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                      battery_capacity: Optional[float], count: int) -> int:
    """ Sets advanced battery system with surplus controller.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    battery_capacity: float or None
        Capacity of the battery in Wh. In case of None default is used.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if battery_capacity is not None:
        my_advanced_battery_config = advanced_battery_bslib.BatteryConfig.get_default_config(
            e_bat_custom=battery_capacity, p_inv_custom=battery_capacity * 0.5 * 1e3, source_weight=count)
    else:
        my_advanced_battery_config = advanced_battery_bslib.BatteryConfig.get_default_config(source_weight=count)
    count += 1
    my_advanced_battery = advanced_battery_bslib.Battery(my_simulation_parameters=my_simulation_parameters, config=my_advanced_battery_config)

    my_electricity_controller.add_component_input_and_connect(
        source_component_class=my_advanced_battery, source_component_output=my_advanced_battery.AcBatteryPower,
        source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.BATTERY, lt.InandOutputType.ELECTRICITY_REAL],
        source_weight=my_advanced_battery.source_weight)

    electricity_to_or_from_battery_target = my_electricity_controller.add_component_output(
        source_output_name=lt.InandOutputType.ELECTRICITY_TARGET, source_tags=[lt.ComponentType.BATTERY, lt.InandOutputType.ELECTRICITY_TARGET],
        source_weight=my_advanced_battery.source_weight, source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT)

    my_advanced_battery.connect_dynamic_input(
        input_fieldname=advanced_battery_bslib.Battery.LoadingPowerInput, src_object=electricity_to_or_from_battery_target)
    my_sim.add_component(my_advanced_battery)

    return count


def configure_water_heating(
        my_sim: Any, my_simulation_parameters: SimulationParameters, my_occupancy: loadprofilegenerator_connector.Occupancy,
        water_heating_system_installed: HeatingSystems, count: int) -> int:
    """ Sets Boiler with Heater, L1 Controller and L2 Controller for Water Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_occupancy: Occupancy
        The initialized occupancy component.
    water_heating_system_installed: str
        Type of installed WaterHeatingSystem
    count: int
        Integer tracking component hierachy for EMS.

    """
    boiler_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_boiler()
    fuel_translator = {HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS, HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
                       HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING}
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_waterheating()
    heater_config.fuel = fuel_translator[water_heating_system_installed]
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw('DHW' + water_heating_system_installed.value)
    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1

    heater_config.power_th = my_occupancy.max_hot_water_demand * (4180 / 3600) * 0.5 \
        * (3600 / my_simulation_parameters.seconds_per_timestep) \
        * (HouseholdWarmWaterDemandConfig.ww_temperature_demand - HouseholdWarmWaterDemandConfig.freshwater_temperature)

    my_boiler = generic_hot_water_storage_modular.HotWaterStorage(my_simulation_parameters=my_simulation_parameters, config=boiler_config)
    my_boiler.connect_only_predefined_connections(my_occupancy)
    my_sim.add_component(my_boiler)

    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                          config=heater_l1_config)
    my_heater_controller_l1.connect_only_predefined_connections(my_boiler)
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)

    my_sim.add_component(my_heater)
    my_sim.add_component(my_heater_controller_l1)

    my_boiler.connect_only_predefined_connections(my_heater)
    return count


def configure_water_heating_electric(
        my_sim: Any, my_simulation_parameters: SimulationParameters,
        my_occupancy: loadprofilegenerator_connector.Occupancy,
        my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
        my_weather: weather.Weather, water_heating_system_installed: HeatingSystems,
        controlable: bool, count: int) -> int:
    """ Sets Boiler with Heater, L1 Controller and L2 Controller for Water Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_occupancy: Occupancy
        The initialized occupancy component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    water_heating_system_installed: str
        Type of installed WaterHeatingSystem
    controlable: bool
        True if control of heating device is smart, False if not.
    count: int
        Integer tracking component hierachy for EMS.

    """

    boiler_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_boiler()

    if water_heating_system_installed == HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_waterheating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw('DHWHeatPumpController')
    elif water_heating_system_installed == HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_waterheating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw('BoilerHeatingController')

    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    heatpump_config.power_th = my_occupancy.max_hot_water_demand * (4180 / 3600) * 0.5 \
        * (3600 / my_simulation_parameters.seconds_per_timestep) \
        * (HouseholdWarmWaterDemandConfig.ww_temperature_demand - HouseholdWarmWaterDemandConfig.freshwater_temperature)

    my_boiler = generic_hot_water_storage_modular.HotWaterStorage(my_simulation_parameters=my_simulation_parameters, config=boiler_config)
    my_boiler.connect_only_predefined_connections(my_occupancy)
    my_sim.add_component(my_boiler)

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                            config=heatpump_l1_config)
    my_heatpump_controller_l1.connect_only_predefined_connections(my_boiler)
    my_sim.add_component(my_heatpump_controller_l1)

    my_heatpump = generic_heat_pump_modular.ModularHeatPump(config=heatpump_config, my_simulation_parameters=my_simulation_parameters)
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)
    my_boiler.connect_only_predefined_connections(my_heatpump)

    if controlable:
        my_heatpump_controller_l1.connect_only_predefined_connections(my_electricity_controller)
        my_heatpump_controller_l1.connect_input(my_heatpump_controller_l1.StorageTemperatureModifier, my_electricity_controller.component_name,
                                                my_electricity_controller.StorageTemperatureModifier)
        my_heatpump.connect_only_predefined_connections(my_electricity_controller)
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)

        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)
    else:
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
                                                                  source_weight=999)
    return count


def configure_heating(my_sim: Any, my_simulation_parameters: SimulationParameters, my_building: building.Building,
                      heating_system_installed: HeatingSystems, count: int) -> Tuple[Component, int]:
    """ Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    heating_system_installed: str
        Type of installed HeatingSystem
    count: int
        Integer tracking component hierachy for EMS.

    """
    fuel_translator = {HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS, HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
                       HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING}
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_heating()
    heater_config.fuel = fuel_translator[heating_system_installed]
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller(heating_system_installed.value)
    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1

    heater_config.power_th = my_building.max_thermal_building_demand_in_watt

    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                          config=heater_l1_config)
    my_heater_controller_l1.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)
    my_sim.add_component(my_heater)

    return my_heater, count


def configure_heating_electric(my_sim: Any, my_simulation_parameters: SimulationParameters, my_building: building.Building,
                               my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                               my_weather: weather.Weather, heating_system_installed: HeatingSystems, controlable: bool,
                               count: int) -> Tuple[Component, int]:
    """ Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem
    controlable: bool
        True if control of heating device is smart, False if not.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if heating_system_installed == HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller('HeatigHeatPumpController')
    elif heating_system_installed == HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller('ElectricHeatingController')

    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    heatpump_config.power_th = my_building.max_thermal_building_demand_in_watt

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                            config=heatpump_l1_config)
    my_heatpump_controller_l1.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_heatpump_controller_l1)

    my_heatpump = generic_heat_pump_modular.ModularHeatPump(config=heatpump_config, my_simulation_parameters=my_simulation_parameters)
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)

    if controlable:
        my_heatpump_controller_l1.connect_input(my_heatpump_controller_l1.StorageTemperatureModifier, my_electricity_controller.component_name,
                                                my_electricity_controller.BuildingTemperatureModifier)
        my_heatpump.connect_only_predefined_connections(my_electricity_controller)
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)

        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)
    else:
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
                                                                  source_weight=999)
    return my_heatpump, count


def configure_heating_with_buffer_electric(my_sim: Any, my_simulation_parameters: SimulationParameters, my_building: building.Building,
                                           my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                                           my_weather: weather.Weather, heating_system_installed: HeatingSystems, buffer_volume: Optional[float],
                                           controlable: bool, count: int) -> Tuple:
    """ Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem.
    buffer_volume: float or None
        Volume of buffer storage in liters. In case of None default is used.
    controlable: bool
        True if control of heating device is smart, False if not.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if heating_system_installed == HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_buffer('BufferHeatPumpController')
    elif heating_system_installed == HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_buffer('BufferElectricHeatingController')

    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    if buffer_volume is not None:
        buffer_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_buffer(buffer_volume)
        buffer_config.power = float(my_building.max_thermal_building_demand_in_watt)

    building_heating_controller_config = controller_l1_building_heating.L1BuildingHeatingConfig.get_default_config_heating("buffer")
    [buffer_config.source_weight, building_heating_controller_config.source_weight] = [count] * 2
    count += 1

    heatpump_config.power_th = my_building.max_thermal_building_demand_in_watt

    my_buffer = generic_hot_water_storage_modular.HotWaterStorage(my_simulation_parameters=my_simulation_parameters, config=buffer_config)
    my_sim.add_component(my_buffer)

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                            config=heatpump_l1_config)
    my_heatpump_controller_l1.connect_only_predefined_connections(my_buffer)
    my_sim.add_component(my_heatpump_controller_l1)
    my_heatpump = generic_heat_pump_modular.ModularHeatPump(config=heatpump_config, my_simulation_parameters=my_simulation_parameters)
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)

    if controlable:
        my_heatpump_controller_l1.connect_only_predefined_connections(my_electricity_controller)
        my_heatpump_controller_l1.connect_input(my_heatpump_controller_l1.StorageTemperatureModifier, my_electricity_controller.component_name,
                                                my_electricity_controller.StorageTemperatureModifier)
        my_heatpump.connect_only_predefined_connections(my_electricity_controller)
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)

        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.ComponentType.HEAT_PUMP, lt.InandOutputType.ELECTRICITY_REAL],
                                                                  source_weight=my_heatpump.source_weight)
    else:
        my_electricity_controller.add_component_input_and_connect(source_component_class=my_heatpump,
                                                                  source_component_output=my_heatpump.ElectricityOutput,
                                                                  source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT,
                                                                  source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
                                                                  source_weight=999)

    my_buffer_controller = controller_l1_building_heating.L1BuildingHeatController(my_simulation_parameters=my_simulation_parameters,
                                                                                   config=building_heating_controller_config)
    my_buffer_controller.connect_only_predefined_connections(my_building)
    my_buffer_controller.connect_only_predefined_connections(my_electricity_controller)
    my_sim.add_component(my_buffer_controller)
    my_buffer.connect_input(my_buffer.L1DeviceSignal,
                            my_buffer_controller.component_name,
                            my_buffer_controller.boiler_signal)
    my_buffer.connect_only_predefined_connections(my_heatpump)
    my_building.add_component_input_and_connect(source_component_class=my_buffer, source_component_output=my_buffer.HeatToBuilding,
                                                source_load_type=lt.LoadTypes.HEATING, source_unit=lt.Units.WATT,
                                                source_tags=[lt.InandOutputType.HEAT_TO_BUILDING], source_weight=count - 1)

    return my_heatpump, my_buffer, count


def configure_heating_with_buffer(my_sim: Any, my_simulation_parameters: SimulationParameters, my_building: building.Building,
                                  heating_system_installed: HeatingSystems, buffer_volume: Optional[float], count: int) -> Tuple:
    """ Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem.
    buffer_volume: float or None
        Volume of buffer storage in liters. In case of None default is used.
    count: int
        Integer tracking component hierachy for EMS.

    """
    fuel_translator = {HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS, HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
                       HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING}
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_heating()
    heater_config.fuel = fuel_translator[heating_system_installed]
    # heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_pump_controller()
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller(heating_system_installed.value)
    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1

    if buffer_volume is not None:
        buffer_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_buffer(buffer_volume)
        buffer_config.power = float(my_building.max_thermal_building_demand_in_watt)

    building_heating_controller_config = controller_l1_building_heating.L1BuildingHeatingConfig.get_default_config_heating("buffer")
    [buffer_config.source_weight, building_heating_controller_config.source_weight] = [count] * 2
    count += 1

    heater_config.power_th = my_building.max_thermal_building_demand_in_watt

    my_buffer = generic_hot_water_storage_modular.HotWaterStorage(my_simulation_parameters=my_simulation_parameters, config=buffer_config)
    my_sim.add_component(my_buffer)
    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(my_simulation_parameters=my_simulation_parameters,
                                                                          config=heater_l1_config)
    my_heater_controller_l1.connect_only_predefined_connections(my_buffer)
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)
    my_sim.add_component(my_heater)

    my_buffer_controller = controller_l1_building_heating.L1BuildingHeatController(my_simulation_parameters=my_simulation_parameters,
                                                                                   config=building_heating_controller_config)
    my_buffer_controller.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_buffer_controller)
    my_buffer.connect_input(my_buffer.L1DeviceSignal,
                            my_buffer_controller.component_name,
                            my_buffer_controller.boiler_signal)
    my_buffer.connect_only_predefined_connections(my_heater)

    my_building.add_component_input_and_connect(source_component_class=my_buffer, source_component_output=my_buffer.HeatToBuilding,
                                                source_load_type=lt.LoadTypes.HEATING, source_unit=lt.Units.WATT,
                                                source_tags=[lt.InandOutputType.HEAT_TO_BUILDING], source_weight=count - 1)

    my_buffer_controller = controller_l1_building_heating.L1BuildingHeatController(my_simulation_parameters=my_simulation_parameters,
                                                                                   config=building_heating_controller_config)
    return my_heater, my_buffer, count


def configure_elctrolysis_h2storage_chp_system(my_sim: Any, my_simulation_parameters: SimulationParameters, my_building: building.Building,
                                               my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
                                               chp_power: Optional[float], h2_storage_size: Optional[float], electrolyzer_power: Optional[float],
                                               count: int) -> Tuple[generic_CHP.GCHP, int]:
    """ Sets electrolysis, H2-storage and chp system.

    Parameters
    ----------
    my_sim: str
        filename of orginal built example.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    chp_power: float or None
        Maximum power (thermal+electrical+loss) of CHP in Watt.
    h2_storage_size: float or None
        Maximum capacity of hydrogen storage in kg hydrogen.
    electrolyzer_power: float or None
        Maximum power of electrolyzer in Watt.
    count: int
        Integer tracking component hierachy for EMS.

    """
    # Fuel Cell default configurations
    l2_config = controller_l2_generic_heat_simple.L2GenericHeatConfig.get_default_config_heating("chp")
    l2_config.source_weight = count
    l1_config_chp = generic_CHP.L1CHPConfig.get_default_config()
    l1_config_chp.source_weight = count
    if chp_power is not None:
        chp_config = generic_CHP.GCHPConfig(name='CHP', source_weight=count, p_el=0.3 * chp_power, p_th=0.5 * chp_power, p_fuel=chp_power)
    else:
        chp_config = generic_CHP.GCHPConfig.get_default_config()
        chp_config.source_weight = count
    count += 1

    # fuel cell
    my_chp = generic_CHP.GCHP(my_simulation_parameters=my_simulation_parameters, config=chp_config)
    my_sim.add_component(my_chp)

    # heat controller of fuel cell
    my_chp_controller_l2 = controller_l2_generic_heat_simple.L2GenericHeatController(my_simulation_parameters=my_simulation_parameters, config=l2_config)
    my_chp_controller_l2.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_chp_controller_l2)

    # run time controller of fuel cell
    my_chp_controller_l1 = generic_CHP.L1GenericCHPRuntimeController(my_simulation_parameters=my_simulation_parameters, config=l1_config_chp)
    my_chp_controller_l1.connect_only_predefined_connections(my_chp_controller_l2)
    my_sim.add_component(my_chp_controller_l1)
    my_chp.connect_only_predefined_connections(my_chp_controller_l1)

    # electricity controller of fuel cell
    my_electricity_controller.add_component_input_and_connect(
        source_component_class=my_chp, source_component_output=my_chp.ElectricityOutput, source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT, source_tags=[lt.ComponentType.FUEL_CELL, lt.InandOutputType.ELECTRICITY_REAL], source_weight=my_chp.source_weight)
    electricity_from_fuelcell_target = my_electricity_controller.add_component_output(
        source_output_name=lt.InandOutputType.ELECTRICITY_TARGET, source_tags=[lt.ComponentType.FUEL_CELL, lt.InandOutputType.ELECTRICITY_TARGET],
        source_weight=my_chp.source_weight, source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT)
    my_chp_controller_l1.connect_dynamic_input(input_fieldname=generic_CHP.L1GenericCHPRuntimeController.ElectricityTarget,
                                               src_object=electricity_from_fuelcell_target)

    # electrolyzer default configuration
    l1_config_electrolyzer = generic_electrolyzer.L1ElectrolyzerConfig.get_default_config()
    l1_config_electrolyzer.source_weight = count
    if electrolyzer_power is not None:
        electrolyzer_config = generic_electrolyzer.GenericElectrolyzerConfig(
            name='Electrolyzer', source_weight=count, min_power=0.5 * electrolyzer_power, max_power=electrolyzer_power,
            min_hydrogen_production_rate_hour=0.125 * electrolyzer_power, max_hydrogen_production_rate_hour=2 * electrolyzer_power)
    else:
        electrolyzer_config = generic_electrolyzer.GenericElectrolyzerConfig.get_default_config()
        electrolyzer_config.source_weight = count
    count += 1

    # electrolyzer
    my_electrolyzer = generic_electrolyzer.GenericElectrolyzer(my_simulation_parameters=my_simulation_parameters, config=electrolyzer_config)
    my_sim.add_component(my_electrolyzer)

    # run time controller of electrolyzer
    my_electrolyzer_controller_l1 = generic_electrolyzer.L1GenericElectrolyzerController(my_simulation_parameters=my_simulation_parameters, config=l1_config_electrolyzer)
    my_sim.add_component(my_electrolyzer_controller_l1)
    my_electrolyzer.connect_only_predefined_connections(my_electrolyzer_controller_l1)

    # electricity controller of fuel cell
    my_electricity_controller.add_component_input_and_connect(
        source_component_class=my_electrolyzer, source_component_output=my_electrolyzer.ElectricityOutput, source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT, source_tags=[lt.ComponentType.ELECTROLYZER, lt.InandOutputType.ELECTRICITY_REAL],
        source_weight=my_electrolyzer.source_weight)
    electricity_to_electrolyzer_target = my_electricity_controller.add_component_output(
        source_output_name=lt.InandOutputType.ELECTRICITY_TARGET, source_tags=[lt.ComponentType.ELECTROLYZER, lt.InandOutputType.ELECTRICITY_TARGET],
        source_weight=my_electrolyzer.source_weight, source_load_type=lt.LoadTypes.ELECTRICITY, source_unit=lt.Units.WATT)
    my_electrolyzer_controller_l1.connect_dynamic_input(
        input_fieldname=generic_electrolyzer.L1GenericElectrolyzerController.l2_ElectricityTarget, src_object=electricity_to_electrolyzer_target)

    if h2_storage_size is not None:
        h2_storage_config = generic_hydrogen_storage.GenericHydrogenStorageConfig.get_default_config(
            capacity=h2_storage_size, max_charging_rate=h2_storage_size * 1e-2, max_discharging_rate=h2_storage_size * 1e-2, source_weight=count)
    else:
        h2_storage_config = generic_hydrogen_storage.GenericHydrogenStorageConfig.get_default_config(source_weight=count)
    my_h2storage = generic_hydrogen_storage.GenericHydrogenStorage(my_simulation_parameters=my_simulation_parameters, config=h2_storage_config)
    my_h2storage.connect_only_predefined_connections(my_electrolyzer)
    my_h2storage.connect_only_predefined_connections(my_chp)
    my_sim.add_component(my_h2storage)

    my_electrolyzer_controller_l1.connect_only_predefined_connections(my_h2storage)
    my_chp_controller_l1.connect_only_predefined_connections(my_h2storage)

    return my_chp, count
